"""
Persistence Module that supports both
local and remote persistence

The PersistentDict offers a dict-like interface with some powerful features:
- Mutability: Changes to the dict are automatically persisted
- Serialization: Values are automatically serialized and deserialized
  - Supports JSON, Pickle, MsgPack, and Custom Serializers
- Compression: Values are automatically compressed and decompressed
    - Supports Zlib Compression
- Remote Persistence: Supports Redis as a remote persistence backend
- Local Persistence: Supports Local Filesystem as a local persistence backend
- Supports async operations


Usage:

    from lazyops.libs.persistence import PersistentDict

    # Create a new PersistentDict
    # with default settings

    cache = PersistentDict("my_cache")

    # Set a key
    cache["foo"] = "bar"
    cache["x"] = 1

    # Get a key
    print(cache["foo"])
    print(cache["x"])

    # Mutate a key
    cache["x"] += 1
    print(cache["x"])

    # Delete a key
    del cache["foo"]
    print(cache["foo"])

    # Check if a key exists
    print("foo" in cache)

    # Get all keys
    print(cache.keys())

    # Get all values
    print(cache.values())

    # Get all items
    print(cache.items())

    # Get the length of the cache
    print(len(cache))

    # Clear the cache
    cache.clear()
 
"""


from .main import PersistentDict
from .backends import (
    LocalStatefulBackend, 
    RedisStatefulBackend, 
    StatefulBackendT,
)
from .serializers import (
    JsonSerializer, PickleSerializer, MsgPackSerializer, BaseSerializer,
    SerializerT, get_serializer
)


from lazyops.imports._sqlalchemy import resolve_sqlalchemy

resolve_sqlalchemy(True)

from sqlalchemy import create_engine, text, inspect
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.row import Row
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.sql.elements import TextClause
try:
    from sqlalchemy.orm.decl_api import DeclarativeBase
except ImportError:
    from sqlalchemy.orm.decl_api import DeclarativeMeta as DeclarativeBase


from __future__ import annotations

import os
import abc
import atexit
import tempfile
import pathlib
import filelock
import contextlib
from lazyops.utils.serialization import Json
from typing import Optional, Dict, Any, Set, List, Union, TYPE_CHECKING



class TemporaryData(abc.ABC):
    """
    Temporary Data that deletes itself on exit
    """
    def __init__(
        self, 
        filepath: Optional[pathlib.Path] = None,
        is_multithreaded: Optional[bool] = False,
    ):
        if not filepath: filepath = pathlib.Path(tempfile.mktemp())
        self.filepath = filepath
        self.filelock_path = filepath.with_suffix('.lock')
        self.filelock = filelock.FileLock(lock_file = self.filelock_path.as_posix(), thread_local = False)
        self.is_multithreaded = is_multithreaded
        with self.filelock:
            if not self.filepath.exists():
                self.filepath.write_text('{}')
        if self.is_multithreaded and not self.get('process_id'):
            self['process_id'] = os.getpid()
        atexit.register(self.cleanup_on_exit)

    @property
    def data(self) -> Dict[str, Any]:
        """
        Returns the data
        """
        with self.filelock:
            return Json.loads(self.filepath.read_text())
        
    @contextlib.contextmanager
    def ctx(self) -> Dict[str, Union[List[Any], Dict[str, Any], Any]]:
        """
        Returns the context
        """
        with self.filelock:
            data = self.data
            try:
                yield data
            finally:
                self.filepath.write_text(Json.dumps(data, indent = 2))
    
    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Returns the value for the given key
        """
        return self.data.get(key, default)
    
    def __contains__(self, key: str) -> bool:
        """
        Returns whether the key is in the data
        """
        return key in self.data
    
    def __getitem__(self, key: str) -> Any:
        """
        Returns the value for the given key
        """
        return self.data.get(key)
    
    def __setitem__(self, key: str, value: Any):
        """
        Sets the value for the given key
        """
        with self.ctx() as data:
            data[key] = value

    def __delitem__(self, key: str):
        """
        Deletes the value for the given key
        """
        with self.ctx() as data:
            del data[key]


    def __iter__(self):
        """
        Returns the iterator
        """
        return iter(self.data)
    
    def __len__(self) -> int:
        """
        Returns the length
        """
        return len(self.data)
    
    def __repr__(self) -> str:
        """
        Returns the representation
        """
        return repr(self.data)
    
    def __str__(self) -> str:
        """
        Returns the string representation
        """
        return str(self.data)
    
    def __bool__(self) -> bool:
        """
        Returns whether the data is empty
        """
        return bool(self.data)
    
    def __eq__(self, other: Any) -> bool:
        """
        Returns whether the data is equal to the other
        """
        return self.data == other
    
    def keys(self) -> Set[str]:
        """
        Returns the keys
        """
        return self.data.keys()
    
    def setdefault(self, key: str, default: Any) -> Any:
        """
        Sets the default value for the given key
        """
        with self.ctx() as data:
            value = data.setdefault(key, default)
        return value
        
    def close(self):
        """
        Closes the filelock
        """
        self.filelock.release()

    def append(self, key: str, value: Any) -> bool:
        """
        Appends the value to the list
        """
        with self.ctx() as data:
            if key not in data:
                data[key] = []
            if value not in data[key]:
                data[key].append(value)
                return False
        return True

    def cleanup_on_exit(self):
        """
        Cleans up on exit
        """
        if not self.filepath.exists() and not self.filelock_path.exists():
            return
        if self.is_multithreaded and self['process_id'] != os.getpid():
            return
        with contextlib.suppress(Exception):
            self.close()
            self.filepath.unlink()
            self.filelock_path.unlink()


    def has_logged(self, key: str) -> bool:
        """
        Returns whether the key has been logged
        """
        return self.append('logged', key)
        
    @classmethod
    def from_module(
        cls, 
        module_name: str, 
        data_dir: Optional[str] = '.data', 
        is_multithreaded: Optional[bool] = False,
        **kwargs
    ) -> TemporaryData:
        """
        Returns a temporary data from the module
        """
        from lazyops.utils.assets import get_module_path
        module_path = get_module_path(module_name)
        module_dir = module_path.joinpath(data_dir)
        module_dir.mkdir(parents = True, exist_ok = True)
        filepath = module_dir.joinpath(f'{module_name}.tmp.json')
        return cls(filepath = filepath, is_multithreaded = is_multithreaded, **kwargs)


import typing
from re import sub

def to_camel_case(text: str):
    """
    Convert a snake str to camel case.
    multi_master -> multiMaster
    """
    components = text.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() for x in components[1:])

def to_snake_case(text: str):
    """
    multiMaster -> multi_master
    """
    return '_'.join(
        sub('([A-Z][a-z]+)', r' \1',
        sub('([A-Z]+)', r' \1',
        text.replace('-', ' '))).split()).lower()


def to_snake_case_key(text: str):
    """
    multi_master -> multi-master
    """
    return '-'.join(
        sub('([A-Z][a-z]+)', r' \1',
        sub('([A-Z]+)', r' \1',
        text.replace('_', ' '))).split()).lower()
        

def to_snake_case_args(text: str):
    """
    multiMaster -> multi-master
    """
    return '-'.join(
        sub('([A-Z][a-z]+)', r' \1',
        sub('([A-Z]+)', r' \1',
        text.replace('-', ' '))).split()).lower()

def to_graphql_format(
    data: typing.Dict
) -> typing.Dict:
    """
    converts all keys in the data to
    camelcase
    """
    result = {}
    for key, value in data.items():
        if isinstance(value, dict):
            result[to_camel_case(key)] = to_graphql_format(value)
        elif isinstance(value, list):
            values = []
            for item in value:
                if isinstance(item, (dict, list)):
                    values.append(to_graphql_format(item))
                else:
                    values.append(item)
            result[to_camel_case(key)] = values
        else:
            result[to_camel_case(key)] = value    
    return result

def from_graphql_format(
    data: typing.Dict
) -> typing.Dict:
    """
    converts all camelcase keys in the data to
    snake_case
    """
    result = {}
    for key, value in data.items():
        if isinstance(value, dict):
            result[to_snake_case(key)] = from_graphql_format(value)
        elif isinstance(value, list):
            values = []
            for item in value:
                if isinstance(item, (dict, list)):
                    values.append(from_graphql_format(item))
                else:
                    values.append(item)
            result[to_snake_case(key)] = values
        else:
            result[to_snake_case(key)] = value    
    return result
        


import contextlib
from sqlalchemy import func, text
from sqlalchemy import delete as sqlalchemy_delete
from sqlalchemy import update as sqlalchemy_update
from sqlalchemy import exists as sqlalchemy_exists
from sqlalchemy.future import select
from sqlalchemy.sql.expression import Select
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import selectinload, joinedload, immediateload
from sqlalchemy import Column, Integer, DateTime, String, Text, ForeignKey, Boolean, Identity, Enum

from typing import Any, Generator, AsyncGenerator, Iterable, Optional, Union, Type, Dict, cast, TYPE_CHECKING, List, Tuple, TypeVar, Callable
from lazyops.utils import create_unique_id, create_timestamp
from lazyops.utils.logs import logger
from lazyops.types import lazyproperty
from lazyops.libs.psqldb.base import Base, PostgresDB, AsyncSession, Session
from lazyops.libs.psqldb.utils import SQLJson, get_pydantic_model, object_serializer, get_sqlmodel_dict
from fastapi.exceptions import HTTPException
from pydantic import BaseModel

ModelType = TypeVar("ModelType", bound = Base)
ORMModelT = TypeVar("ORMModelT", bound = 'ORMType')

_attr_func_map = {
    "selectinload": selectinload,
    "joinedload": joinedload,
    "immediateload": immediateload,
}

def get_attr_func(
    attr: Union[str, Callable] = None,
):
    if attr and isinstance(attr, str):
        return _attr_func_map.get(attr, selectinload)
    return attr or selectinload


class ORMType(object):

    id: Any = None

    @classmethod
    def _filter(cls, query: Optional[Select] = None, **kwargs) -> Select:
        """
        Build a filter query
        """
        query = query if query is not None else select(cls)
        return query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])

    def _filter_update_data(self, **kwargs) -> Optional[Dict[str, Any]]:
        """
        Filter update data
        """
        data = {}
        for field, value in kwargs.items():
            if not hasattr(self, field):
                continue
            if getattr(self, field) == value:
                continue
            data[field] = value
        return data or None

    @classmethod
    def _handle_exception(cls, msg: Optional[str] = None, e: Optional[Exception] = None):
        """
        Handle exception
        """
        msg = msg or f"{cls.__class__.__name__} not found"
        logger.trace(msg, error = e)
        raise HTTPException(status_code = 404, detail = msg)
    

    @classmethod
    def create(cls, **kwargs) -> ORMModelT:
        """
        Create a new instance of the model
        """
        new_cls = cls(**kwargs)
        with PostgresDB.session() as db_sess:
            db_sess.add(new_cls)
            db_sess.commit()
            db_sess.refresh(new_cls)
        return new_cls
    
    @classmethod
    async def acreate(cls, **kwargs) -> 'ORMType':
        """
        Create a new instance of the model
        """
        async with PostgresDB.async_session() as db_sess:
            new_cls = cls(**kwargs)
            db_sess.add(new_cls)
            await db_sess.commit()
            await db_sess.refresh(new_cls)
        return new_cls
    
    # @classmethod
    # def update(cls, id: Any, **kwargs):
    #     with PostgresDB.session() as db_sess:
    #         query = (
    #             sqlalchemy_update(cls)
    #             .where(cls.id == id)
    #             .values(**kwargs)
    #             .execution_options(synchronize_session="fetch")
    #         )
    #         db_sess.execute(query)
    #         db_sess.commit()
    
    # @classmethod
    # async def aupdate(cls, id: Any, **kwargs):
    #     async with PostgresDB.async_session() as db_sess:


    #         query = (
    #             sqlalchemy_update(cls)
    #             .where(cls.id == id)
    #             .values(**kwargs)
    #             .execution_options(synchronize_session="fetch")
    #         )
    #         await db_sess.execute(query)
    #         await db_sess.commit()

    def update(self, **kwargs) -> 'ORMType':
        """
        Update a record
        """
        filtered_update = self._filter_update_data(**kwargs)
        if not filtered_update:
            return self
        with PostgresDB.session() as db_sess:
            for field, value in filtered_update.items():
                setattr(self, field, value)
                db_sess.add(self)
                db_sess.commit()
                db_sess.refresh(self)
        return self

    async def aupdate(self, **kwargs) -> 'ORMType':
        """
        Update a record
        """
        filtered_update = self._filter_update_data(**kwargs)
        if not filtered_update:
            return self
        async with PostgresDB.async_session() as db_sess:
            for field, value in filtered_update.items():
                setattr(self, field, value)
                db_sess.add(self)
                await db_sess.commit()
                await db_sess.refresh(self)
        return self



    @classmethod
    def get(
        cls, 
        # value: Any, 
        # key: str = "id", 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        **kwargs,
    ) -> 'ORMType':
        """
        Get a record
        """
        with PostgresDB.session(ro=readonly) as db_sess:
            # query = select(cls).where(getattr(cls, key) == value)
            query = cls._filter(**kwargs)
            if load_attrs:
                load_attr_method = get_attr_func(load_attr_method)
                for attr in load_attrs:
                    query = query.options(load_attr_method(getattr(cls, attr)))
            results = db_sess.execute(query)
            (result,) = results.one()
            if not result and raise_exceptions:
                cls._handle_exception()
            return result

    @classmethod
    async def aget(
        cls, 
        # value: Any, 
        # key: str = "id",
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        **kwargs,
    ) -> Type['ORMType']:
        """
        Get a record
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            # query = select(cls).where(*[getattr(cls, key) == value for key, value in kwargs.items()])
            query = cls._filter(**kwargs)
            if load_attrs:
                load_attr_method = get_attr_func(load_attr_method)
                for attr in load_attrs:
                    query = query.options(load_attr_method(getattr(cls, attr)))
            results = await db_sess.execute(query)
            (result,) = results.one()
            if not result and raise_exceptions:
                cls._handle_exception()
            return result

    @classmethod
    def get_all(
        cls,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        **kwargs,
    ) -> List['ORMType']:
        """
        Get all records
        """
        with PostgresDB.session(ro=readonly) as db_sess:
            # query = select(cls)
            query = cls._filter(**kwargs)
            if load_attrs:
                load_attr_method = get_attr_func(load_attr_method)
                for attr in load_attrs:
                    query = query.options(load_attr_method(getattr(cls, attr)))
            results = db_sess.execute(query)
            return results.scalars().all()
    
    @classmethod
    async def aget_all(
        cls,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        **kwargs,
    ) -> List['ORMType']:
        """
        Get all records
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            # query = select(cls)
            # if kwargs:
            #     query = query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])
            query = cls._filter(**kwargs)
            if load_attrs:
                load_attr_method = get_attr_func(load_attr_method)
                for attr in load_attrs:
                    query = query.options(load_attr_method(getattr(cls, attr)))
            results = await db_sess.execute(query)
            return results.scalars().all()
    
    @classmethod
    def delete(cls, id: Any):
        with PostgresDB.session() as db_sess:
            query = sqlalchemy_delete(cls).where(cls.id == id)
            db_sess.execute(query)
            db_sess.commit()
    
    @classmethod
    async def async_delete(cls, **kwargs) -> ModelType:
        """
        Delete a record
        """
        obj = await cls.aget(**kwargs, raise_exceptions = True)
        async with PostgresDB.async_session() as db_sess:
            db_sess.delete(obj)
            await db_sess.commit()
        return obj
    
    @classmethod
    def delete_all(cls):
        with PostgresDB.session() as db_sess:
            query = sqlalchemy_delete(cls)
            db_sess.execute(query)
            db_sess.commit()
    
    @classmethod
    async def adelete_all(cls):
        async with PostgresDB.async_session() as db_sess:
            query = sqlalchemy_delete(cls)
            await db_sess.execute(query)
            await db_sess.commit()
    
    @classmethod
    async def aexists(cls, **kwargs) -> bool:
        """
        Return True if a record exists
        """
        async with PostgresDB.async_session(ro=True) as db_sess:
            query = cls._filter(sqlalchemy_exists(), **kwargs).select()
            # result = await db_sess.execute(query)
            # return bool(result.scalar())
            return bool(await db_sess.scalar(query))


    @classmethod
    def count(
        cls, 
        # owner_id: Optional[Any] = None,
        # owner_key: Optional[str] = "owner_id",
        readonly: Optional[bool] = True,
        **kwargs
    ) -> int:
        """
        Count the number of records in the table.
        """
        query = select(func.count(cls.id)).filter(*[getattr(cls, key) == value for key, value in kwargs.items()]) \
            if kwargs else select(func.count(cls.id))

        try:
            with PostgresDB.session(ro=readonly) as db_sess:
                return db_sess.scalar(query)
        except Exception as e:
            logger.error(e)
            return 0
        
    @classmethod
    async def acount(
        cls, 
        # owner_id: Optional[Any] = None,
        # owner_key: Optional[str] = "owner_id",
        readonly: Optional[bool] = True,
        **kwargs
    ) -> int:
        """
        Count the number of records in the table.
        """
        query = cls._filter(**kwargs) \
            if kwargs else select(func.count(cls.id))
        # query = select(func.count(cls.id)).filter(getattr(cls, owner_key) == owner_id) \
        #     if kwargs else select(func.count(cls.id))
        try:
            async with PostgresDB.async_session(ro=readonly) as db_sess:
                return await db_sess.scalar(query)
        except Exception as e:
            logger.error(e)
            return 0

    
    def dict(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        safe_encode: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        data = self.pydantic_model.dict(
            exclude=exclude,
            include=include,
            **kwargs
        )
        if safe_encode: data = {key: object_serializer(value) for key, value in data.items()}
        return data
    
    def json(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude_none: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        return self.pydantic_model.json(
            exclude=exclude,
            include=include,
            exclude_none=exclude_none,
            **kwargs
        )

    def diff(
        self,
        other: Union[Any, 'ORMType']
    ) -> Dict[str, Any]:
        """
        Return a dictionary of the differences between this model and another.
        """
        return {
            key: value
            for key, value in self.dict().items()
            if key in other.dict() and self.dict()[key] != other.dict()[key]
        }
    
    @lazyproperty
    def pydantic_model(self) -> Type[BaseModel]:
        """
        Return the Pydantic model for this ORM model.
        """
        return get_pydantic_model(self)

    @classmethod
    def filter_first(
        cls, 
        attr: str, 
        value: Any,
        model_type: Optional[ModelType] = None,
        readonly: Optional[bool] = False,
    ) -> Optional[ModelType]:
        model_type = model_type or cls
        attr_obj = getattr(model_type, attr)
        stmt = select(model_type).where(attr_obj == value)
        with PostgresDB.session(ro=readonly) as db_sess:
            return cls.first(db_sess, stmt, model_type)

    @classmethod
    async def afilter_first(
        cls, 
        attr: str, 
        value: Any,
        model_type: Optional[ModelType] = None,
        readonly: Optional[bool] = False,
    ) -> Optional[ModelType]:
        """
        Return the first result of a query.
        """
        model_type = model_type or cls
        attr_obj = getattr(model_type, attr)
        stmt = select(model_type).where(attr_obj == value)
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            return await cls.afirst(db_sess, stmt, model_type)

    @classmethod
    def first(
        cls, 
        db: Session, 
        stmt: Select,
        model_type: ModelType,
    ) -> Optional[ModelType]:
        """
        Return the first result of a query.
        """
        res = (db.scalars(stmt)).first()
        res = cast(model_type, res)
        return res
    
    @classmethod
    async def afirst(
        cls, 
        db: AsyncSession, 
        stmt: Select,
        model_type: ModelType,
    ) -> Optional[ModelType]:
        """
        Return the first result of a query.
        """
        res = (await db.scalars(stmt)).first()
        res = cast(model_type, res)
        return res
    
    @classmethod
    async def aget_or_create(cls, _defaults: Dict[str, Any] = None, **kwargs) -> Tuple[ModelType, bool]:
        """
        Create a new instance of the model, or return the existing one.
        """
        try:
            return await cls.aget(**kwargs), False
        except NoResultFound:
            _defaults = _defaults or {}
            return await cls.acreate(**{**kwargs, **_defaults}), True
            
    @classmethod
    async def aget_or_none(cls, **kwargs) -> Optional[ModelType]:
        """
        Return an instance of the model, or None.
        """
        try:
            return await cls.get(cls, **kwargs)
        except NoResultFound:
            return None
    
    @classmethod
    async def aget_or_create_or_update(cls, _defaults: Dict[str, Any] = None, **kwargs) -> Tuple[ModelType, bool]:
        """
        Create a new instance of the model, 
        or return the existing one after updating it.

        Returns a tuple of (instance, created | updated), where created is a boolean
        """
        try:
            res = await cls.aget(**kwargs)
            return (await res.aupdate(**_defaults), True) \
                if _defaults else (res, False)
        except NoResultFound:
            _defaults = _defaults or {}
            return await cls.acreate(**{**kwargs, **_defaults}), False

    @classmethod
    def get_by_owner(
        cls,
        owner_id: int,
        owner_key: str = "owner_id",
        name: Optional[str] = None,
        name_key: Optional[str] = "name",
        skip: int = 0, 
        limit: int = 100,
        order_by: Optional[Any] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
    ):
        """
        Fetch all records by owner_id
        """
        with PostgresDB.session(ro=readonly) as db_sess:
            query = select(cls).where(getattr(cls, owner_key) == owner_id)
            if name:
                query = query.where(getattr(cls, name_key) == name)
            if load_attrs:
                load_attr_method = get_attr_func(load_attr_method)
                for attr in load_attrs:
                    query = query.options(load_attr_method(getattr(cls, attr)))
            if skip is not None:
                query = query.offset(skip)
            if limit:
                query = query.limit(limit)
            if order_by is not None:
                query = query.order_by(order_by)
            results = db_sess.execute(query)
            return results.scalars().all()
    
    @classmethod
    async def aget_by_owner(
        cls,
        owner_id: int,
        owner_key: str = "owner_id",
        name: Optional[str] = None,
        name_key: Optional[str] = "name",
        skip: int = 0, 
        limit: int = 100,
        order_by: Optional[Any] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly: Optional[bool] = False,
    ):
        """
        Fetch all records by owner_id
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            query = select(cls).where(getattr(cls, owner_key) == owner_id)
            if name:
                query = query.where(getattr(cls, name_key) == name)
            if load_attrs:
                load_attr_method = get_attr_func(load_attr_method)
                for attr in load_attrs:
                    query = query.options(load_attr_method(getattr(cls, attr)))
            if skip is not None:
                query = query.offset(skip)
            if limit:
                query = query.limit(limit)
            if order_by is not None:
                query = query.order_by(order_by)
            results = await db_sess.execute(query)
            return results.scalars().all()


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.dict()})>"

"""
Async ORM Type
"""

# AsyncORMTypeT = TypeVar("AsyncORMTypeT", "AsyncORMType", Base)


# class AsyncORMType(object):

class AsyncORMType(object):
    """
    Only contains async methods
    """

    id: str = Column(Text, default = create_unique_id, primary_key = True, index = True)
    created_at = Column(DateTime(timezone=True), default = create_timestamp, server_default = func.now())
    updated_at = Column(DateTime(timezone=True), default = create_timestamp, server_default = func.now(), onupdate = create_timestamp)


    @classmethod
    def _filter(
        cls, 
        query: Optional[Select] = None, 
        **kwargs
    ) -> Select:
        """
        Build a filter query
        """
        query = query if query is not None else select(cls)
        return query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])

    @classmethod
    def _build_query(
        cls,
        query: Optional[Select] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        **kwargs
    ) -> Select:
        """
        Build a query
        """
        query = query if query is not None else select(cls)
        query = query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])
        if load_attrs:
            load_attr_method = get_attr_func(load_attr_method)
            for attr in load_attrs:
                query = query.options(load_attr_method(getattr(cls, attr)))
        return query

    def _filter_update_data(self, **kwargs) -> Optional[Dict[str, Any]]:
        """
        Filter update data
        """
        data = {}
        for field, value in kwargs.items():
            if not hasattr(self, field): continue
            if getattr(self, field) == value: 
                # logger.warning(f"Skipping {field}={getattr(self, field)} as value is the same: {value}")
                continue
            data[field] = value
        return data or None

    @classmethod
    def _handle_exception(cls, msg: Optional[str] = None, e: Optional[Exception] = None, verbose: Optional[bool] = False):
        """
        Handle exception
        """
        msg = msg or f"{cls.__name__} not found"
        if verbose: logger.trace(msg, error = e)
        raise e or HTTPException(status_code = 404, detail = msg)
        # raise HTTPException(status_code = 404, detail = msg)

    
    @classmethod
    async def create(cls, **kwargs) -> Type['AsyncORMType']:
        """
        Create a new instance of the model
        """
        async with PostgresDB.async_session() as db_sess:
            new_cls = cls(**kwargs)
            db_sess.add(new_cls)
            await db_sess.flush()
            await db_sess.commit()
            # await db_sess.flush(new_cls)
            await db_sess.refresh(new_cls)
        return new_cls


    @classmethod
    def _create(cls, **kwargs) -> Type['AsyncORMType']:
        """
        Create a new instance of the model
        """
        with PostgresDB.session() as db_sess:
            new_cls = cls(**kwargs)
            db_sess.add(new_cls)
            db_sess.flush()
            db_sess.commit()
            db_sess.refresh(new_cls)
            
        return new_cls

    async def refresh(self, session: Optional[AsyncSession] = None, **kwargs):
        """
        Refresh a record
        """
        # async with PostgresDB.async_session(session = session) as db_sess:
        #     query = sqlalchemy_update(self.__class__).where(
        #         self.__class__.id == self.id)
        #     self = (await db_sess.execute(query)).scalar_one_or_none()
            # await db_sess.expire(self)
            # local_object = await db_sess.merge(self)
            # db_sess.add(local_object)
            # await db_sess.commit()
            # await db_sess.refresh(local_object)

        # self = local_object
        # return local_object
    
        async with PostgresDB.async_session() as db_sess:
            # await db_sess.expire(self)
            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            await db_sess.commit()
            await db_sess.refresh(local_object)

            self = local_object
        # return local_object
    
    async def _update(self, **kwargs):
        """
        Update a record
        """
        async with PostgresDB.async_session() as db_sess:
            query = sqlalchemy_update(self.__class__).where(
                self.__class__.id == self.id).values(
                **kwargs
            )
            await db_sess.execute(query)
            await db_sess.commit()

            # refresh attributes
        self = await self.get(id = self.id)
        # return self
        # await db_sess.refresh(self)

    
    async def update_inplace(self, **kwargs):
        """
        Update a record inplace
        """
        async with PostgresDB.async_session() as db_sess:
            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            await db_sess.refresh(local_object)
            await db_sess.commit()
            await db_sess.flush()
            self = local_object
        return self
    
    async def save(self, **kwargs):
        """
        Save a record
        """
        async with PostgresDB.async_session() as db_sess:
            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            await db_sess.refresh(local_object)
            await db_sess.commit()
        
        self = local_object
        return self

    async def update(self, **kwargs) -> Type['AsyncORMType']:
        """
        Update a record
        """
        filtered_update = self._filter_update_data(**kwargs)
        if not filtered_update: return self
        async with PostgresDB.async_session() as db_sess:
            for field, value in filtered_update.items():
                setattr(self, field, value)

            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            await db_sess.refresh(local_object)
            await db_sess.commit()
        
        self = local_object
        return self


    @classmethod
    async def get(
        cls, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        _model_type: Optional[ModelType] = None,
        _verbose: Optional[bool] = False,
        **kwargs,
    ) -> Type['AsyncORMType']:
        """
        Get a record
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            result = (await db_sess.execute(query)).scalar_one_or_none()
            if not result and raise_exceptions:
                cls._handle_exception(e = NoResultFound(), verbose = _verbose)
            if result is not None and _model_type is not None: result = cast(_model_type, result)
        return result
    
    @classmethod
    def _get(
        cls, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        _model_type: Optional[ModelType] = None,
        _verbose: Optional[bool] = False,
        **kwargs,
    ) -> Type['AsyncORMType']:
        """
        Get a record
        """
        with PostgresDB.session(ro=readonly) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            result = (db_sess.execute(query)).scalar_one_or_none()
            if not result and raise_exceptions:
                cls._handle_exception(e = NoResultFound(), verbose = _verbose)
            if result is not None and _model_type is not None: result = cast(_model_type, result)
        return result

    @classmethod
    async def get_all(
        cls,
        skip: Optional[int] = None, 
        limit: Optional[int] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        _model_type: Optional[ModelType] = None,
        **kwargs,
    ) -> List[Type['AsyncORMType']]:
        """
        Get all records
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            if skip is not None: query = query.offset(skip)
            if limit is not None: query = query.limit(limit)
            results = (await db_sess.scalars(query)).all()
            if not results and raise_exceptions:
                cls._handle_exception(e = NoResultFound())
            if results is not None and _model_type is not None: results = cast(List[_model_type], results)
        return results


    @classmethod
    async def first(
        cls, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly: Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        _model_type: Optional[ModelType] = None,
        **kwargs
    ) -> Optional[Union[Type['AsyncORMType'], ModelType]]:
        """
        Return the first result of a query.
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            result = (await db_sess.scalars(query)).first()
            if not result and raise_exceptions:
                cls._handle_exception(e = NoResultFound())
            if result is not None and _model_type is not None: result = cast(_model_type, result)
        return result
    
    
    @classmethod
    async def delete(cls, **kwargs) -> Type['AsyncORMType']:
        """
        Delete a record
        """
        obj = await cls.get(**kwargs, raise_exceptions = True)
        async with PostgresDB.async_session() as db_sess:
            db_sess.delete(obj)
            await db_sess.commit()
        return obj
    
    @classmethod
    async def delete_all(cls):
        async with PostgresDB.async_session() as db_sess:
            query = sqlalchemy_delete(cls)
            await db_sess.execute(query)
            await db_sess.commit()
    
    @classmethod
    async def exists(cls, **kwargs) -> bool:
        """
        Return True if a record exists
        """
        async with PostgresDB.async_session(ro=True) as db_sess:
            query = cls._filter(sqlalchemy_exists(), **kwargs).select()
            res = await db_sess.scalar(query)
            # logger.info(f"{cls.__name__} Exists: {res} | {kwargs}")
        return res
        # return bool(await db_sess.scalar(query))

        
    @classmethod
    async def count(
        cls, 
        readonly: Optional[bool] = True,
        **kwargs
    ) -> int:
        """
        Count the number of records in the table.
        """
        query = cls._filter(**kwargs) \
            if kwargs else select(func.count(cls.id))
        try:
            async with PostgresDB.async_session(ro=readonly) as db_sess:
                return await db_sess.scalar(query)
        except Exception as e:
            logger.error(e)
            return 0

    
    def dict(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        safe_encode: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        data = self.pydantic_model.dict(
            exclude=exclude,
            include=include,
            **kwargs
        )
        if safe_encode: data = {key: object_serializer(value) for key, value in data.items()}
        return data
    
    def json(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude_none: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        return self.pydantic_model.json(
            exclude=exclude,
            include=include,
            exclude_none=exclude_none,
            **kwargs
        )

    def diff(
        self,
        other: Union[Any, 'AsyncORMType']
    ) -> Dict[str, Any]:
        """
        Return a dictionary of the differences between this model and another.
        """
        return {
            key: value
            for key, value in self.dict().items()
            if key in other.dict() and self.dict()[key] != other.dict()[key]
        }
    
    @lazyproperty
    def pydantic_model(self) -> Type[BaseModel]:
        """
        Return the Pydantic model for this ORM model.
        """
        return get_pydantic_model(self)

    
    @classmethod
    async def get_or_create(
        cls, 
        filterby: Optional[Iterable[str]] = None,
        **kwargs
    ) -> Tuple[Type['AsyncORMType'], bool]:
        """
        Create a new instance of the model, or return the existing one.
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        result = await cls.get(**_filterby, raise_exceptions = False, _verbose = False)
        if result is not None: return result, False
        return await cls.create(**kwargs), True
        # try:
        #     return await cls.get(**kwargs), False
        # except NoResultFound:
        #     _defaults = _defaults or {}
        #     return await cls.create(**{**kwargs, **_defaults}), True

    @classmethod
    def register(
        cls, 
        filterby: Optional[Iterable[str]] = None,
        **kwargs
    ) -> Tuple[Type['AsyncORMType'], bool]:
        """
        Create a new instance of the model, or return the existing one.

        Non-async version of `get_or_create` for use in synchronous code.
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        result = cls._get(**_filterby, raise_exceptions = False, _verbose = False)
        return (result, False) if result is not None else (cls._create(**kwargs), True)
            
    @classmethod
    async def get_or_none(cls, **kwargs) -> Optional[Type['AsyncORMType']]:
        """
        Return an instance of the model, or None.
        """
        return await cls.get(**kwargs, raise_exceptions = False, _verbose = False)
        # try:
        #     return await cls.get(cls, **kwargs)
        # except NoResultFound:
        #     return None
    
    @classmethod
    async def get_or_create_or_update(
        cls, 
        filterby: Optional[Iterable[str]] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        **kwargs: Dict
    ) -> Tuple[Type['AsyncORMType'], bool]:
        """
        Create a new instance of the model, 
        or return the existing one after updating it.

        filterby: A list of fields to filter by when checking for an existing record.

        Returns a tuple of (instance, created | updated), where created is a boolean
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        result = await cls.get(**_filterby, raise_exceptions = False, _verbose = False, load_attrs = load_attrs, load_attr_method = load_attr_method)
        if result is None: 
            new_cls = await cls.create(**kwargs)

            return new_cls, True
        if update_data := result._filter_update_data(**kwargs):
            return (await result.update(**update_data), True)
        return (result, False)
    
        
    @classmethod
    async def create_or_update(
        cls, 
        filterby: Optional[Iterable[str]] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        **kwargs
    ) -> Tuple[Type['AsyncORMType'], bool]:
        """
        Create a new instance of the model, 
        or return the existing one after updating it.

        Returns a tuple of (instance, (true: created or updated, false: not updated), where created is a boolean
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        # if await cls.exists(**_filterby):
        result = await cls.get(**_filterby, raise_exceptions = False, _verbose = False, load_attrs = load_attrs, load_attr_method = load_attr_method)
        if result is None: 
            return await cls.create(**kwargs), True
        if update_data := result._filter_update_data(**kwargs):
            logger.info(f"Updating {result} with {update_data}")
            return (await result.update(**update_data), True)
        return (result, False)
    

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.dict()})>"
    


    
# SQLModelT = Type['SQLModel']
SQLModelT = TypeVar('SQLModelT', bound = 'SQLModel')

class SQLModel(Base):
    """
    Abstract base class for all SQL models.
    """
    __abstract__ = True
    __allow_unmapped__ = True
    
    
    id: str = Column(Text, default = create_unique_id, primary_key = True, index = True, unique = True)
    created_at = Column(DateTime(timezone=True), default = create_timestamp, server_default = func.now())
    updated_at = Column(DateTime(timezone=True), default = create_timestamp, server_default = func.now(), onupdate = create_timestamp)

    """
    Core Methods
    """

    @classmethod
    async def create(
        cls, 
        session: Optional[AsyncSession] = None,
        **kwargs
    ) -> SQLModelT:
        """
        Create a new instance of the model
        """
        async with PostgresDB.async_session(session = session) as db_sess:
            new_cls = cls(**kwargs)
            db_sess.add(new_cls)
            await db_sess.flush()
            await db_sess.commit()
            await db_sess.refresh(new_cls)
        return new_cls

    @classmethod
    def _create(
        cls: SQLModelT, 
        session: Optional[Session] = None,
        **kwargs
    ) -> SQLModelT:
        """
        Create a new instance of the model
        """
        with PostgresDB.session(session = session) as db_sess:
            new_cls = cls(**kwargs)
            db_sess.add(new_cls)
            db_sess.flush()
            db_sess.commit()
            db_sess.refresh(new_cls)
            
        return new_cls
    

    async def refresh(
        self, 
        session: Optional[AsyncSession] = None,
        **kwargs
    ):
        """
        Refresh a record
        """
        async with PostgresDB.async_session(session = session) as db_sess:
            # await db_sess.expire(self)
            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            await db_sess.commit()
            await db_sess.refresh(local_object)

        self = local_object
        return local_object
    
    async def _update(
        self, 
        session: Optional[AsyncSession] = None,
        **kwargs
    ):
        """
        Update a record
        """
        async with PostgresDB.async_session(session = session) as db_sess:
            query = sqlalchemy_update(self.__class__).where(
                self.__class__.id == self.id).values(
                **kwargs
            )
            await db_sess.execute(query)
            await db_sess.commit()

            # refresh attributes
        self = await self.get(id = self.id)
        # return self
        # await db_sess.refresh(self)

    
    async def update_inplace(
        self, 
        session: Optional[AsyncSession] = None,
        **kwargs
    ):
        """
        Update a record inplace
        """
        async with PostgresDB.async_session(session = session) as db_sess:
            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            # await db_sess.refresh(local_object)
            await db_sess.commit()
            await db_sess.flush()
            self = local_object
        return self
    
    async def save(self, **kwargs):
        """
        Save a record
        """
        async with PostgresDB.async_session() as db_sess:
            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            # logger.info(f"Saving {local_object}")
            # await db_sess.refresh(local_object)
            await db_sess.commit()
            self = local_object
        return self

    async def update(self, session: Optional[AsyncSession] = None, **kwargs) -> Type['AsyncORMType']:
        """
        Update a record
        """
        filtered_update = self._filter_update_data(**kwargs)
        if not filtered_update: return self
        async with PostgresDB.async_session(session = session) as db_sess:
            for field, value in filtered_update.items():
                setattr(self, field, value)

            local_object = await db_sess.merge(self)
            db_sess.add(local_object)
            await db_sess.refresh(local_object)
            # if session is None: 
            await db_sess.commit()
        
            self = local_object
        return self


    @classmethod
    async def get(
        cls: SQLModelT, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        session: Optional[AsyncSession] = None,
        _model_type: Optional[ModelType] = None,
        _verbose: Optional[bool] = False,
        **kwargs,
    ) -> SQLModelT:
        """
        Get a record
        """
        async with PostgresDB.async_session(ro=readonly, session = session) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            result = (await db_sess.execute(query)).scalar_one_or_none()
            if not result and raise_exceptions:
                cls._handle_exception(e = NoResultFound(), verbose = _verbose)
            if result is not None and _model_type is not None: result = cast(_model_type, result)
        return result
    

    @classmethod
    async def get_only_one(
        cls: SQLModelT, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        session: Optional[AsyncSession] = None,
        _model_type: Optional[ModelType] = None,
        _verbose: Optional[bool] = False,
        **kwargs,
    ) -> SQLModelT:
        """
        Get a record
        """
        async with PostgresDB.async_session(ro=readonly, session = session) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            query = query.order_by(cls.created_at.desc())
            # result = (await db_sess.execute(query)).scalar_one_or_none()
            result = (await db_sess.scalars(query)).all()
            # logger.warning(f"Found {len(result)} results for {cls.__name__} with {kwargs}: {result}")
            if not result and raise_exceptions:
                cls._handle_exception(e = NoResultFound(), verbose = _verbose)
            if len(result) > 1:
                # Keep the latest one
                kept_result = result[0]
                if _verbose: logger.warning(f"Found {len(result)} results for {cls.__name__} with {kwargs}. Keeping {kept_result.id}")
                # Delete the rest
                for res in result[1:]:
                    await db_sess.delete(res)
                # await db_sess.flush()
                await db_sess.commit()
                result = kept_result
            elif result:
                result = result[0]
            else:
                result = None
            if result is not None and _model_type is not None: result = cast(_model_type, result)
        return result
    
    @classmethod
    def _get(
        cls: SQLModelT, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        session: Optional[Session] = None,
        _model_type: Optional[ModelType] = None,
        _verbose: Optional[bool] = False,
        **kwargs,
    ) -> SQLModelT:
        """
        Get a record
        """
        with PostgresDB.session(ro=readonly, session = session) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            result = (db_sess.execute(query)).scalar_one_or_none()
            if not result and raise_exceptions:
                cls._handle_exception(e = NoResultFound(), verbose = _verbose)
            if result is not None and _model_type is not None: result = cast(_model_type, result)
        return result

    @classmethod
    async def get_all(
        cls: SQLModelT,
        skip: Optional[int] = None, 
        limit: Optional[int] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly : Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        _model_type: Optional[ModelType] = None,
        **kwargs,
    ) -> List[SQLModelT]:
        """
        Get all records
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            if skip is not None: query = query.offset(skip)
            if limit is not None: query = query.limit(limit)
            results = (await db_sess.scalars(query)).all()
            if not results and raise_exceptions:
                cls._handle_exception(e = NoResultFound())
            if results is not None and _model_type is not None: results = cast(List[_model_type], results)
        return results


    @classmethod
    async def first(
        cls, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        readonly: Optional[bool] = False,
        raise_exceptions: Optional[bool] = True,
        _model_type: Optional[ModelType] = None,
        **kwargs
    ) -> Optional[Union[SQLModelT, ModelType]]:
        """
        Return the first result of a query.
        """
        async with PostgresDB.async_session(ro=readonly) as db_sess:
            query = cls._build_query(load_attrs = load_attrs, load_attr_method = load_attr_method, **kwargs)
            result = (await db_sess.scalars(query)).first()
            if not result and raise_exceptions:
                cls._handle_exception(e = NoResultFound())
            if result is not None and _model_type is not None: result = cast(_model_type, result)
        return result
    
    
    @classmethod
    async def delete(cls, **kwargs) -> SQLModelT:
        """
        Delete a record
        """
        obj = await cls.get(**kwargs, raise_exceptions = True)
        async with PostgresDB.async_session() as db_sess:
            await db_sess.delete(obj)
            await db_sess.commit()
        return obj
    
    @classmethod
    async def delete_all(cls):
        async with PostgresDB.async_session() as db_sess:
            query = sqlalchemy_delete(cls)
            await db_sess.execute(query)
            await db_sess.commit()
    
    @classmethod
    async def exists(cls, session: Optional[AsyncSession] = None, **kwargs) -> bool:
        """
        Return True if a record exists
        """
        async with PostgresDB.async_session(ro=True, session = session) as db_sess:
            query = cls._filter(sqlalchemy_exists(), **kwargs).select()
            res = await db_sess.scalar(query)
            # logger.info(f"{cls.__name__} Exists: {res} | {kwargs}")
        return res
        # return bool(await db_sess.scalar(query))

        
    @classmethod
    async def count(
        cls, 
        readonly: Optional[bool] = True,
        **kwargs
    ) -> int:
        """
        Count the number of records in the table.
        """
        query = cls._filter(**kwargs) \
            if kwargs else select(func.count(cls.id))
        try:
            async with PostgresDB.async_session(ro=readonly) as db_sess:
                return await db_sess.scalar(query)
        except Exception as e:
            logger.error(e)
            return 0

    
    def dict(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        safe_encode: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        data = self.pydantic_model.dict(
            exclude=exclude,
            include=include,
            **kwargs
        )
        if safe_encode: data = {key: object_serializer(value) for key, value in data.items()}
        return data
    
    def json(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude_none: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        return self.pydantic_model.json(
            exclude=exclude,
            include=include,
            exclude_none=exclude_none,
            **kwargs
        )

    def diff(
        self,
        other: Union[Any, 'SQLModel'],
    ) -> Dict[str, Any]:
        """
        Return a dictionary of the differences between this model and another.
        """
        return {
            key: value
            for key, value in self.dict().items()
            if key in other.dict() and self.dict()[key] != other.dict()[key]
        }
    
    def sqlmodel_dict(
        self, 
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude_none: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        model_dict = get_sqlmodel_dict(self)
        if exclude is not None: model_dict = {key: value for key, value in model_dict.items() if key not in exclude}
        if include is not None: model_dict = {key: value for key, value in model_dict.items() if key in include}
        if exclude_none: model_dict = {key: value for key, value in model_dict.items() if value is not None}
        return model_dict

    @lazyproperty
    def pydantic_model(self) -> Type[BaseModel]:
        """
        Return the Pydantic model for this ORM model.
        """
        return get_pydantic_model(self)

    
    @classmethod
    async def get_or_create(
        cls, 
        filterby: Optional[Iterable[str]] = None,
        session: Optional[AsyncSession] = None,
        **kwargs
    ) -> Tuple[SQLModelT, bool]:
        """
        Create a new instance of the model, or return the existing one.
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        # async with PostgresDB.async_session(session = session) as db_sess:
        result = await cls.get(
            raise_exceptions = False, 
            _verbose = False, 
            session = session,
            **_filterby, 
        )
        if result is not None: return result, False
        return await cls.create(
            session = session, 
            **kwargs
        ), True
    

    @classmethod
    def register(
        cls: SQLModelT, 
        filterby: Optional[Iterable[str]] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        session: Optional[Session] = None,
        **kwargs
    ) -> Tuple[SQLModelT, bool]:
        """
        Create a new instance of the model, or return the existing one.

        Non-async version of `get_or_create` for use in synchronous code.
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        with PostgresDB.session(session = session) as db_sess:
            result = cls._get(**_filterby, raise_exceptions = False, _verbose = False, session = db_sess, load_attrs = load_attrs, load_attr_method = load_attr_method)
            return (result, False) if result is not None else (cls._create(session = db_sess, **kwargs), True)
            
    @classmethod
    async def get_or_none(cls: SQLModelT, session: Optional[AsyncSession] = None, **kwargs) -> Optional[SQLModelT]:
        """
        Return an instance of the model, or None.
        """
        return await cls.get(**kwargs, session = session, raise_exceptions = False, _verbose = False)
        # try:
        #     return await cls.get(cls, **kwargs)
        # except NoResultFound:
        #     return None
    
    @classmethod
    async def get_or_create_or_update(
        cls: SQLModelT, 
        filterby: Optional[Iterable[str]] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        session: Optional[AsyncSession] = None,
        _only_new: Optional[bool] = False,
        **kwargs: Dict
    ) -> Tuple[SQLModelT, bool]:
        """
        Create a new instance of the model, 
        or return the existing one after updating it.

        filterby: A list of fields to filter by when checking for an existing record.

        Returns a tuple of (instance, created | updated), where created is a boolean
        """
        # filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        filterby = list(kwargs.keys()) if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        # async with PostgresDB.async_session(session = session) as db_sess:
        # result = await cls.get(
        result = await cls.get_only_one(
            load_attrs = load_attrs,
            load_attr_method = load_attr_method,
            session = session, 
            raise_exceptions = False, 
            _verbose = False,
            **_filterby, 
        )
        if result is None: 
            new_cls = await cls.create(
                session = session,
                **kwargs, 
            )
            return new_cls, True
        if update_data := result._filter_update_data(**kwargs):
            return (
                await result.update(
                    session = session, 
                    **update_data
                ), (False if _only_new else True))
        return (result, False)
    
        
    @classmethod
    async def create_or_update(
        cls: SQLModelT, 
        filterby: Optional[Iterable[str]] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        session: Optional[AsyncSession] = None,
        _only_new: Optional[bool] = False,
        **kwargs
    ) -> Tuple[SQLModelT, bool]:
        """
        Create a new instance of the model, 
        or return the existing one after updating it.

        Returns a tuple of (instance, (true: created or updated, false: not updated), where created is a boolean
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        # async with PostgresDB.async_session(session = session) as db_sess:
        result = await cls.get(
            **_filterby, 
            load_attrs = load_attrs,
            load_attr_method = load_attr_method,
            session = session, 
            raise_exceptions = False, 
            _verbose = False
        )
        if result is None: 
            return await cls.create(
                session = session, 
                **kwargs
            ), True
        if update_data := result._filter_update_data(**kwargs):
            logger.info(f"Updating {result} with {update_data}")
            return (
                await result.update(
                session = session, 
                **update_data
                ), (False if _only_new else True)
            )
        return (result, False)
    

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.dict()})>"
    
    def _supdate(
        self, 
        session: Optional[Session] = None,
        **kwargs
    ):
        """
        Update a record
        """
        with PostgresDB.session(session = session) as db_sess:
            query = sqlalchemy_update(self.__class__).where(
                self.__class__.id == self.id).values(
                **kwargs
            )
            db_sess.execute(query)
            db_sess.commit()

            # refresh attributes
        return self._get(id = self.id)


    def supdate(self, session: Optional[Session] = None, **kwargs):
        """
        Update a record
        """
        filtered_update = self._filter_update_data(**kwargs)
        if not filtered_update: return self
        with PostgresDB.session(session = session) as db_sess:
            for field, value in filtered_update.items():
                setattr(self, field, value)

            local_object = db_sess.merge(self)
            db_sess.add(local_object)
            db_sess.refresh(local_object)
            db_sess.commit()
            self = local_object
        return self

    @classmethod
    def query(
        cls: SQLModelT,
        query: str,
        session: Optional[Session] = None,
        **kwargs
    ) -> Any:
        """
        Executes a text query
        """
        with PostgresDB.session(session = session) as db_sess:
            query = query.replace('[table]', cls.__tablename__)
            stmt = text(query)
            result = db_sess.execute(stmt, kwargs)
            db_sess.commit()
        return result

    @classmethod
    async def aquery(
        cls: SQLModelT,
        query: str,
        session: Optional[AsyncSession] = None,
        **kwargs
    ) -> Any:
        """
        Executes a text query
        """
        async with PostgresDB.async_session(session = session) as db_sess:
            query = query.replace('[table]', cls.__tablename__)
            stmt = text(query)
            result = await db_sess.execute(stmt, kwargs)
            await db_sess.commit()
        return result

    @classmethod
    def _get_or_create(
        cls: SQLModelT, 
        filterby: Optional[Iterable[str]] = None,
        session: Optional[Session] = None,
        **kwargs
    ) -> Tuple[SQLModelT, bool]:
        """
        Create a new instance of the model, or return the existing one.
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        # async with PostgresDB.async_session(session = session) as db_sess:
        result = cls._get(
            raise_exceptions = False, 
            _verbose = False, 
            session = session,
            **_filterby, 
        )
        if result is not None: return result, False
        return cls._create(
            session = session, 
            **kwargs
        ), True

    @classmethod
    def _get_or_create_or_update(
        cls: SQLModelT, 
        filterby: Optional[Iterable[str]] = None,
        session: Optional[Session] = None,
        _only_new: Optional[bool] = False,
        **kwargs: Dict
    ) -> Tuple[SQLModelT, bool]:
        """
        Create a new instance of the model, 
        or return the existing one after updating it.

        filterby: A list of fields to filter by when checking for an existing record.

        Returns a tuple of (instance, created | updated), where created is a boolean
        """
        filterby = [list(kwargs.keys())[0]] if filterby is None else filterby
        _filterby = {key: kwargs.get(key) for key in filterby}
        # async with PostgresDB.async_session(session = session) as db_sess:
        result = cls._get(
            session = session, 
            raise_exceptions = False, 
            _verbose = False,
            **_filterby, 
        )
        if result is None: 
            new_cls = cls._create(
                session = session,
                **kwargs, 
            )
            return new_cls, True
        if update_data := result._filter_update_data(**kwargs):
            return (
                result._supdate(
                    session = session, 
                    **update_data
                ), (False if _only_new else True))
        return (result, False)

    @classmethod
    def _exists(cls, session: Optional[Session] = None, **kwargs) -> bool:
        """
        Return True if a record exists
        """
        with PostgresDB.session(ro=True, session = session) as db_sess:
            query = cls._filter(sqlalchemy_exists(), **kwargs).select()
            res = db_sess.scalar(query)
        return res
    
    @classmethod
    def _get_or_none(cls, session: Optional[Session] = None, **kwargs) -> Optional[SQLModelT]:
        """
        Return an instance of the model, or None.
        """
        return cls._get(**kwargs, session = session, raise_exceptions = False, _verbose = False)

    """
    Utility Methods
    """
    
    @classmethod
    def _filter(
        cls: SQLModelT, 
        query: Optional[Select] = None, 
        **kwargs
    ) -> Select:
        """
        Build a filter query
        """
        query = query if query is not None else select(cls)
        return query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])

    @classmethod
    def _nested_kwarg_parser(
        cls: SQLModelT,
        top_level_obj: Optional[SQLModelT] = None,
        **kwargs: Dict[str, Any],
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Parses nested dict kwargs into a flat dict
        """
        attrs = []
        for key, value in kwargs.items():
            obj_attr = getattr(cls, key, None) if top_level_obj is None else getattr(top_level_obj, key, None)
            if obj_attr is None: continue
            if isinstance(value, dict):
                attrs.append(obj_attr.has(**value))
                # attrs.extend(cls._nested_kwarg_parser(top_level_obj, **value))
            else:
                attrs.append(obj_attr == value)
        logger.warning(f'{attrs}')
        return attrs

    @classmethod
    def _build_query(
        cls: SQLModelT,
        query: Optional[Select] = None,
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        **kwargs
    ) -> Select:
        """
        Build a query
        """
        query = query if query is not None else select(cls)
        query = query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])
        # query = query.where(*cls._nested_kwarg_parser(**kwargs))
        if load_attrs:
            load_attr_method = get_attr_func(load_attr_method)
            for attr in load_attrs:
                query = query.options(load_attr_method(getattr(cls, attr)))
        return query

    def _filter_update_data(self, **kwargs) -> Optional[Dict[str, Any]]:
        """
        Filter update data
        """
        data = {}
        for field, value in kwargs.items():
            if not hasattr(self, field): continue
            with contextlib.suppress(ValueError):
                if hasattr(value, 'all') and hasattr(getattr(self, field), 'all') \
                    and getattr(self, field).all() == value.all():
                    continue

                if getattr(self, field) == value: 
                    # logger.warning(f"Skipping {field}={getattr(self, field)} as value is the same: {value}")
                    continue
            
            data[field] = value
        return data or None

    @classmethod
    def _handle_exception(
        cls, 
        msg: Optional[str] = None, 
        e: Optional[Exception] = None, 
        verbose: Optional[bool] = False
    ):
        """
        Handle exception
        """
        msg = msg or f"{cls.__name__} not found"
        if verbose: logger.trace(msg, error = e)
        raise e or HTTPException(status_code = 404, detail = msg)

    




"""
Generic helpers for apis when working with multi workers
"""

from lazyops.imports._psutil import resolve_psutil

resolve_psutil(True)

import os
import psutil
from typing import List, Optional
from lazyops.types import BaseModel, lazyproperty


class NodeProcess(BaseModel):
    pid: int # Current Worker PID
    name: str # Process Name
    exe: str # Command Line
    cmdline: List[str]

    is_worker: Optional[bool] = None
    has_workers: Optional[bool] = None

    parent: Optional[psutil.Process] = None
    linked: Optional[List[psutil.Process]] = []

    # Context vars
    runtime: Optional[str] = None
    entrypoint: Optional[str] = None
    exact_match: Optional[bool] = None
    worker_entrypoint: Optional[str] = None


    @lazyproperty
    def parent_pid(self) -> Optional[int]:
        """
        Parent PID
        """
        return self.parent.pid if self.parent else None
    
    @lazyproperty
    def parent_cmdline(self) -> Optional[List[str]]:
        """
        Parent Command Line
        """
        return self.parent.cmdline() if self.parent else None
    
    @lazyproperty
    def parent_exe(self) -> Optional[str]:
        """
        Parent Executable
        """
        return self.parent.exe() if self.parent else None

    @lazyproperty
    def linked_pids(self) -> Optional[List[int]]:
        """
        Linked PID's
        """
        return [p.pid for p in self.linked] if self.linked else None

    @lazyproperty
    def is_parent(self) -> bool:
        """
        If the parent process is none, we assume it is the current
        process. This is the case when running in a single worker
        """
        return (
            self.pid == self.parent_pid \
            if self.parent_pid is not None else True
        )

    @lazyproperty
    def is_multiprocess(self) -> bool:
        """
        If the linked pids are empty, we assume it is a single
        process. This is the case when running in a single worker
        """
        return any('multiprocessing' in cmd for cmd in self.cmdline)

    @staticmethod
    def get_stripped_cmdline(
        cmdline: List[str],
        has_workers: bool = False,
    ) -> str:
        """
        Strips the current process from the cmdline
        and removes params that are not needed
        """
        cmd = ""
        for cmd_arg in cmdline:
            if 'pipe_handle' in cmd_arg:
                cmd_arg = cmd_arg.split(', pipe_handle', 1)[0] + ')'
                # will fix this later
                if has_workers and 'tracker_fd' in cmd_arg:
                    cmd_arg = '(' + cmd_arg.split('(tracker_fd', 1)[0] + ')'
            cmd += f'|{cmd_arg}' 
        return cmd

    @lazyproperty
    def stripped_cmdline(self) -> List[str]:
        """
        Strips the current process from the cmdline
        and removes params that are not needed
        """
        return self.get_stripped_cmdline(self.cmdline)
        # return self.get_stripped_cmdline(self.cmdline, has_workers = self.has_workers)
    
    @lazyproperty
    def total_processes(self) -> int:
        """
        Returns the total number of processes
        """
        return len(self.linked_pids) + 1 if self.linked_pids else 1
    
    @lazyproperty
    def sorted_linked_pids(self) -> List[int]:
        """
        Returns the sorted linked pids
        """
        if not self.linked_pids: return []
        pids = list(self.linked_pids)
        if not self.is_parent:
            pids.append(self.pid)
        if self.parent_pid and self.parent_pid in pids:
            pids.remove(self.parent_pid)
        return sorted(pids, key=int)

    # @lazyproperty
    @property
    def leader_pid(self) -> Optional[int]:
        """
        Returns the leader pid
        """
        if self.is_worker: return self.sorted_linked_pids[1] if len(self.sorted_linked_pids) > 1 else None
        return self.parent_pid if self.is_parent else self.sorted_linked_pids[0]
    
    # @lazyproperty
    @property
    def is_leader(self) -> bool:
        """
        If the current process is the leader

        detect the first worker pid by numerical order
        """
        return True if self.is_parent else self.leader_pid == self.pid
    
    
    @classmethod
    def get_process(
        cls,
        pid: Optional[int] = None,
        runtime: Optional[str] = "uvicorn",
        entrypoint: Optional[str] = "main:app",
        exact_match: Optional[bool] = False,
        is_worker: Optional[bool] = False,
        worker_entrypoint: Optional[str] = None,
        has_workers: Optional[bool] = False,
    ) -> 'NodeProcess':
        """
        Utility Function to get the current process information
        of the current node / worker

        can probably improve the logic but it works for now.
        """
        node_pid = pid or os.getpid()
        node_process = psutil.Process(node_pid)

        process = cls(
            pid = node_pid,
            name = node_process.name(),
            exe = node_process.exe(),
            cmdline = node_process.cmdline(),
            has_workers = has_workers,
            is_worker = is_worker,

            # Context vars
            runtime = runtime,
            entrypoint = entrypoint,
            exact_match = exact_match,
            worker_entrypoint = worker_entrypoint,
        )

        for p in psutil.process_iter(attrs=["pid", "name", "exe", "cmdline"]):
            if p.info["pid"] == node_pid:
                continue

            # Check that the process name is identical to the node
            if p.info['exe'] != process.exe:
                continue

            if not is_worker and \
                any(runtime in cmd_arg for cmd_arg in p.info["cmdline"]) and \
                    any((entrypoint == cmd_arg if exact_match else entrypoint in cmd_arg) for cmd_arg in p.info["cmdline"]):
                
                process.linked.append(p)
                process.parent = p
                continue


            if (
                is_worker
                and worker_entrypoint
                and all(runtime not in cmd_arg for cmd_arg in p.info["cmdline"])
                and any(
                    (
                        worker_entrypoint == cmd_arg
                        if exact_match
                        else worker_entrypoint in cmd_arg
                    )
                    for cmd_arg in p.info["cmdline"]
                )
                and not process.parent
            ):
                process.parent = p
                continue


            # if process.is_multiprocess and process.stripped_cmdline == cls.get_stripped_cmdline(p.info["cmdline"], has_workers = has_workers):
            if process.is_multiprocess and process.stripped_cmdline == cls.get_stripped_cmdline(p.info["cmdline"]):
                process.linked.append(p)
                continue
        
        return process
    
    def reconfigure(self):
        """
        Reconfigures the current process
        """
        self = self.get_process(
            pid = self.pid,
            runtime = self.runtime,
            entrypoint = self.entrypoint,
            exact_match = self.exact_match,
            is_worker = self.is_worker,
            worker_entrypoint = self.worker_entrypoint,
        )



import time

from sqlalchemy import event
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.asyncio import AsyncEngine

from typing import Union, Optional

# https://docs.sqlalchemy.org/en/14/faq/connections.html#faq-execute-retry
# Retryable ReadOnly Event

def reconnecting_engine(
    engine: Union[Engine, AsyncEngine],
    num_retries: Optional[int] = 5, 
    retry_interval: Optional[float] = 10.0,
):
    def _run_with_retries(fn, context, cursor_obj, statement, *arg, **kw):
        for retry in range(num_retries + 1):
            try:
                fn(cursor_obj, statement, context=context, *arg)
            except engine.dialect.dbapi.Error as raw_dbapi_err:
                connection = context.root_connection
                if engine.dialect.is_disconnect(raw_dbapi_err, connection, cursor_obj):
                    if retry > num_retries:
                        raise
                    engine.logger.error(
                        "disconnection error, retrying operation",
                        exc_info=True,
                    )
                    connection.invalidate()

                    # use SQLAlchemy 2.0 API if available
                    if hasattr(connection, "rollback"):
                        connection.rollback()
                    else:
                        trans = connection.get_transaction()
                        if trans:
                            trans.rollback()

                    time.sleep(retry_interval)
                    context.cursor = cursor_obj = connection.connection.cursor()
                else:
                    raise
            else:
                return True
        
    
    e = engine.execution_options(isolation_level="AUTOCOMMIT")
    engine_obj = e if isinstance(engine, Engine) else e.sync_engine
    

    @event.listens_for(engine_obj, "do_execute_no_params")
    def do_execute_no_params(cursor_obj, statement, context):
        return _run_with_retries(
            context.dialect.do_execute_no_params, context, cursor_obj, statement
        )

    @event.listens_for(engine_obj, "do_execute")
    def do_execute(cursor_obj, statement, parameters, context):
        return _run_with_retries(
            context.dialect.do_execute, context, cursor_obj, statement, parameters
        )

    return e


import time
import uuid
import typing
import random
import inspect
import functools
import datetime
import itertools
import asyncio
import contextlib
import async_lru
import signal


from frozendict import frozendict
from typing import Dict, Callable, List, Any, Union, Coroutine, TypeVar, TYPE_CHECKING
from lazyops.utils.logs import default_logger

from lazyops.utils.serialization import (
    object_serializer, object_deserializer,
    ObjectEncoder, ObjectDecoder,
    Json,
)

# For backwards compatibility
from lazyops.utils.lazy import (
    import_string,
    import_function, 
    get_obj_class_name,
    fetch_property,
    is_coro_func,
    lazy_function,
    lazy_import,
)

if TYPE_CHECKING:
    from lazyops.types import BaseModel


def timer(t: typing.Optional[float] = None, msg: typing.Optional[str] = None, logger = default_logger):
    if not t: return time.perf_counter()
    done_time = time.perf_counter() - t
    if msg: logger.info(f'{msg} in {done_time:.2f} secs')
    return done_time

# def timed(func: typing.Callable):
#     """
#     Decorator to time a function
#     """
#     _func_name = func.__name__
#     @functools.wraps(func)
#     async def fx(*args, **kwargs):
#         start = time.perf_counter()
#         if inspect.iscoroutinefunction(func): result = await func(*args, **kwargs)
#         else: result = func(*args, **kwargs)
#         end = time.perf_counter()
#         default_logger.info(f'{_func_name}: {end - start:.4f} secs')
#         return result
#     return fx


def timed(verbose: typing.Optional[bool] = False):
    """
    Decorator to time a function
    """
    def decorator(func):
        if inspect.iscoroutinefunction(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                start = time.perf_counter()
                result = await func(*args, **kwargs)
                end = time.perf_counter()
                if verbose: default_logger.info(f"took {end - start:.2f} secs with args: {args} and kwargs: {kwargs}", prefix = f"{func.__module__}.|g|{func.__name__}|e|", colored = True)
                else: default_logger.info(f"took {end - start:.2f} secs", prefix = f"{func.__module__}.|g|{func.__name__}|e|", colored = True)
                return result
            return wrapper
        
        else:        
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                start = time.perf_counter()
                result = func(*args, **kwargs)
                end = time.perf_counter()
                if verbose: default_logger.info(f"took {end - start:.2f} secs with args: {args} and kwargs: {kwargs}", prefix = f"{func.__module__}.|g|{func.__name__}|e|", colored = True)
                else: default_logger.info(f"took {end - start:.2f} secs", prefix = f"{func.__module__}.|g|{func.__name__}|e|", colored = True)
                return result
            return wrapper
    
    return decorator


def merge_dicts(x: Dict, y: Dict):
    z = x.copy()
    z.update(y)
    return z

def exponential_backoff(
    attempts: int,
    base_delay: int = 1,
    max_delay: int = None,
    jitter: bool = True,
):
    """
    Get the next delay for retries in exponential backoff.

    attempts: Number of attempts so far
    base_delay: Base delay, in seconds
    max_delay: Max delay, in seconds. If None (default), there is no max.
    jitter: If True, add a random jitter to the delay
    """
    if max_delay is None:
        max_delay = float("inf")
    backoff = min(max_delay, base_delay * 2 ** max(attempts - 1, 0))
    if jitter:
        backoff = backoff * random.random()
    return backoff

def retryable(limit: int = 3, delay: int = 3):
    def decorator(func: Callable):
        if not inspect.iscoroutinefunction(func):
            def sync_wrapper(*args, **kwargs):
                for n in range(limit - 1):
                    with contextlib.suppress(Exception):
                        return func(*args, **kwargs)
                    time.sleep(exponential_backoff(n, base_delay=delay))
                return func(*args, **kwargs)
            return sync_wrapper
        else:
            async def async_wrapper(*args, **kwargs):
                for n in range(limit-1):
                    with contextlib.suppress(Exception):
                        return await func(*args, **kwargs)
                    await asyncio.sleep(exponential_backoff(n, base_delay=delay))
                return await func(*args, **kwargs)
            return async_wrapper
    return decorator


T = TypeVar('T')

def get_batches_from_generator(iterable: typing.Iterable[T], n: int) -> typing.Generator[typing.List[T], None, None]:
    """
    Batch elements of an iterable into fixed-length chunks or blocks.
    """
    it = iter(iterable)
    while x := tuple(itertools.islice(it, n)):
        yield x

def split_into_batches(items: List[T], n: int) -> typing.Iterable[typing.List[T]]:
    """
    Splits the items into n amount of equal items

    >>> list(split_into_batches(range(11), 3))
    [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10]]
    """
    n = min(n, len(items))
    k, m = divmod(len(items), n)
    return (items[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n))

def split_into_batches_of_n(iterable: typing.Iterable[T], n: int) -> typing.Iterable[typing.List[T]]:
    """
    Splits the items into fixed-length chunks or blocks.

    >>> list(split_into_batches_of_n(range(11), 3))
    [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]
    """
    return list(get_batches_from_generator(iterable, n))


def split_into_n_batches(iterable: typing.Iterable[T], size: int) -> typing.Iterable[typing.List[T]]:
    """
    Splits the items into n amount of equal items

    >>> list(split_into_batches_of_size(range(11), 3))
    [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]
    """
    return split_into_batches(iterable, size)

def build_batches(iterable: typing.Iterable[T], size: int, fixed_batch_size: bool = True) -> typing.Iterable[typing.List[T]]:
    """
    Builds batches of a given size from an iterable.
    """
    if fixed_batch_size:
        return split_into_batches_of_n(iterable, size)
    return split_into_n_batches(iterable, size)


def split_into_batches_with_index(iterable: typing.Iterable[T], n: int, start: typing.Optional[int] = None) -> typing.Iterable[typing.Tuple[int, List[T]]]:
    """
    Splits the items into fixed-length chunks or blocks.

    >>> list(split_into_batches_of_n(range(11), 3))
    [(0, [0, 1, 2]), (1, [3, 4, 5]), (2, [6, 7, 8]), (3, [9, 10])]
    """
    return list(enumerate(get_batches_from_generator(iterable, n), start = start or 0))

def create_unique_id(
    method: typing.Optional[str] = 'uuid4',
    alph_only: typing.Optional[bool] = False,
    length: typing.Optional[int] = None,
):
    """
    Creates a unique id
    args:
        method: uuid4, uuid1, uuid5, timestamp, secret
        alph_only: if True, returns a string of only alphabets
        length: if specified, returns a string of the specified length
    """
    meth = getattr(uuid, method, None)
    if not meth:  raise ValueError(f'Invalid UUID method: {method}')
    val = str(meth())
    if alph_only: val = ''.join([c for c in val if c.isalpha()])
    if length:
        while len(val) < length:
            val += str(meth())
            if alph_only: val = ''.join([c for c in val if c.isalpha()])
            # remove trailing hyphen
            if val.endswith('-'): val = val[:-1]
        val = val[:length]
    return val

def create_timestamp(
    tz: typing.Optional[datetime.tzinfo] = datetime.timezone.utc,
    as_str: typing.Optional[bool] = False,
):
    """
    Creates a timestamp
    args:
        tz: timezone
        as_str: if True, returns a string
    """
    dt = datetime.datetime.now(tz =tz)
    return dt.isoformat() if as_str else dt

def create_secret(
    nbytes: typing.Optional[int] = 16,    
):
    import secrets
    return secrets.token_hex(nbytes)


def suppress(
    *exceptions: typing.Optional[typing.Union[typing.Tuple[Exception], Exception]],
):
    """
    Wrapper for suppressing exceptions
    args:
        exceptions: exceptions to suppress
    """
    if not exceptions: exceptions = (Exception,)
    def wrapper(func):
        if is_coro_func(func):
            @functools.wraps(func)
            async def wrapped_func(*args, **kwargs):
                with contextlib.suppress(*exceptions):
                    return await func(*args, **kwargs)
            return wrapped_func
        else:
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                with contextlib.suppress(*exceptions):
                    return func(*args, **kwargs)
        return wrapped_func
    return wrapper

"""
Timed Caches
"""

def recursive_freeze(value):
    if not isinstance(value, dict):
        return value
    for k,v in value.items():
        value[k] = recursive_freeze(v)
    return frozendict(value)

# To unfreeze
def recursive_unfreeze(value):
    if isinstance(value, frozendict):
        value = dict(value)
        for k,v in value.items():
            value[k] = recursive_unfreeze(v)
    
    return value

def freeze_args_and_kwargs(*args, **kwargs):
    args = tuple(
        recursive_freeze(arg) if isinstance(arg, dict) else arg
        for arg in args
    )
    kwargs = {k: recursive_freeze(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
    return args, kwargs


def timed_cache(
    secs: typing.Optional[int] = 60 * 60, 
    maxsize: int = 1024,
    invalidate_cache_key: typing.Optional[str] = '_invalidate_cache',
):
    """
    Wrapper for creating a expiring cached function
    args:
        secs: number of seconds to cache the function
        maxsize: maxsize of the cache
        invalidate_cache_key: key to invalidate the cache
    """
    if secs is None: secs = 60
    def wrapper_cache(func):
        if is_coro_func(func):

            @async_lru.alru_cache(maxsize = maxsize, ttl = secs)
            async def _wrapped(*args, **kwargs):
                return await func(*args, **kwargs)
            
            @functools.wraps(func)
            async def wrapped_func(*args, **kwargs):
                _invalidate = kwargs.pop(invalidate_cache_key, None) if invalidate_cache_key else None
                args, kwargs = freeze_args_and_kwargs(*args, **kwargs)
                if _invalidate is True:
                    _wrapped.cache_invalidate(*args, **kwargs)
                # print(f'wrapped: {_wrapped.cache_info()}')
                return await _wrapped(*args, **kwargs)
            
            return wrapped_func
        
        else:
            func = functools.lru_cache(maxsize=maxsize)(func)
            func.lifetime = datetime.timedelta(seconds=secs)
            func.expiration = create_timestamp() + func.lifetime
            def _check_cache(func, invalidate: typing.Optional[bool] = None):
                if invalidate is True or create_timestamp() >= func.expiration:
                    func.cache_clear()
                    func.expiration = create_timestamp() + func.lifetime
            
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                _check_cache(func, invalidate = kwargs.pop(invalidate_cache_key, None) if invalidate_cache_key else None)
                args, kwargs = freeze_args_and_kwargs(*args, **kwargs)
                return func(*args, **kwargs)
            
            return wrapped_func

    return wrapper_cache
        


def with_fail_after(delay: float):
    """
    Creates a timeout for a function
    """

    def wrapper(func):
        @functools.wraps(func)
        def time_limited(*args, **kwargs):
            # Register an handler for the timeout
            def handler(signum, frame):
                raise TimeoutError(f"Timeout for function '{func.__name__}'")

            # Register the signal function handler
            signal.signal(signal.SIGALRM, handler)

            # Define a timeout for your function
            signal.alarm(delay)

            result = None
            try:
                result = func(*args, **kwargs)
            except Exception as exc:
                raise exc
            finally:
                # disable the signal alarm
                signal.alarm(0)

            return result

        return time_limited
    return wrapper


@contextlib.contextmanager
def fail_after(delay: Union[int, float]):
    """
    Creates a timeout for a function
    """
    def signal_handler(signum, frame):
        raise TimeoutError(f"Timed out after {delay}s")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(int(delay))
    try:
        yield
    finally:
        signal.alarm(0)


async def run_as_coro(
    func: Callable,
    *args,
    **kwargs
) -> Any:
    """
    Runs a function as a coroutine
    """
    from .pooler import ThreadPooler
    try:
        return await ThreadPooler.asyncish(func, *args, **kwargs)
    except Exception as e:
        default_logger.trace(f'Error running as coro', error = e)
        raise e


_background_tasks = set()


def create_background_task(func: Union[Callable, Coroutine], *args, **kwargs):
    """
    Creates a background task and adds it to the global set of background tasks
    """
    if inspect.isawaitable(func):
        task = asyncio.create_task(func)
    else:
        task = asyncio.create_task(
            run_as_coro(
                func,
                *args,
                **kwargs
            )
        )
    _background_tasks.add(task)
    task.add_done_callback(_background_tasks.discard)
    return task


async def adeferred_task(
    func: Union[Callable, Coroutine],
    *args,
    wait_time: typing.Optional[float] = 10.0,
    **kwargs,
):
    """
    Runs a function as a background task
    """
    await asyncio.sleep(wait_time)
    return create_background_task(func, *args, **kwargs)



def build_dict_from_str(
    data: str,
    **kwargs
) -> Union[List[Any], Dict[str, Any]]:
    """
    Helper to build a dictionary from a string
    """
    import json
    if (data.startswith('[') and data.endswith(']')) or (data.startswith('{') and data.endswith('}')):
        return json.loads(data)
    return build_dict_from_list(data.split(','), **kwargs)


def build_dict_from_list(
    data: List[str],
    seperator: str = '=',
) -> Dict[str, Any]:
    """
    Builds a dictionary from a list of strings
    """
    import json
    return json.loads(str(dict([item.split(seperator) for item in data])))




class AsyncTaskException(Exception):
    def __init__(self, message, func_name):
        super().__init__(message)
        self.func_name = func_name

    def __str__(self):
        return f"An exception occured in '{self.func_name}': {super().__str__()}"


"""
File Streaming Support
"""

from io import IOBase, TextIOWrapper
from typing import Iterable, Any, Union, Type, Optional

class FileLikeObject(IOBase):

    def __init__(
        self, 
        data: Iterable[Union[bytes, Any]],
        base: Type = bytes,
        # async_enabled: Optional[bool] = False,
        **kwargs,
    ):
        super().__init__()
        self.base: Type[bytes] = base
        self.chunk = base()
        self.offset = 0
        self.iterator = iter(data)
    
    def up_to_iter(self, size: int) -> Iterable[Union[bytes, Any]]:
        """
        Yield up to size bytes from the iterator.
        """
        while size:
            if self.offset == len(self.chunk):
                try: self.chunk = next(self.iterator)
                except StopIteration: break
                else: self.offset = 0
            to_yield = min(size, len(self.chunk) - self.offset)
            self.offset = self.offset + to_yield
            size -= to_yield
            yield self.chunk[self.offset - to_yield : self.offset]

    def readable(self):
        """
        Return True if the stream can be read from. If False, read() will raise
        """
        return True
    
    def writable(self):
        """
        Return True if the stream supports writing. If False, write() will raise
        """
        return False
    
    def read(self, size: int = -1) -> bytes:
        """
        Read and return up to size bytes. If the argument is omitted, None, or
        """
        return self.base().join(
            self.up_to_iter(float('inf') if size is None or size < 0 else size)
        )
    
    def read1(self, size: int = -1) -> bytes:
        """
        Read and return up to size bytes, with at most one call to the underlying
        """
        return self.read(size)
    
    def as_line_iterator(self, newline: Optional[str] = '', encoding: Optional[str] = 'utf-8', **kwargs) -> Iterable[str]:
        """
        Return an iterator the yields lines from the stream.
        """
        return TextIOWrapper(self, newline=newline, encoding=encoding, **kwargs)


class AsyncFileLikeObject(IOBase):

    def __init__(
        self, 
        data: Iterable[Union[bytes, Any]],
        base: Type = bytes,
        **kwargs,
    ):
        super().__init__()
        self.base: Type[bytes] = base
        self.chunk = base()
        self.offset = 0
        self.iterator = aiter(data)
    
    async def up_to_iter(self, size: int) -> Iterable[Union[bytes, Any]]:
        """
        Yield up to size bytes from the iterator.
        """
        while size:
            if self.offset == len(self.chunk):
                try: self.chunk = await anext(self.iterator)
                except StopIteration: break
                else: self.offset = 0
            to_yield = min(size, len(self.chunk) - self.offset)
            self.offset += to_yield
            size -= to_yield
            yield self.chunk[self.offset - to_yield : self.offset]

    def readable(self):
        """
        Return True if the stream can be read from. If False, read() will raise
        """
        return True
    
    def writable(self):
        """
        Return True if the stream supports writing. If False, write() will raise
        """
        return False
    
    async def read(self, size: int = -1) -> bytes:
        """
        Read and return up to size bytes. If the argument is omitted, None, or
        """
        b = []
        async for chunk in self.up_to_iter(float('inf') if size is None or size < 0 else size):
            b += chunk        
        return self.base().join(b) 

    


from __future__ import annotations

"""
The Base Class for Convertio Handlers
"""

import abc
import tempfile
from pathlib import Path, PurePath
from lazyops.utils.pooler import ThreadPoolV2
from lazyops.utils.logs import default_logger, null_logger, Logger
from typing import Any, Union, Optional, Type, Iterable, Callable, Dict, List, Tuple, TypeVar, TYPE_CHECKING

try:
    from xxhash import xxh64 as hash_func
except ImportError:
    from hashlib import md5 as hash_func

try:
    from fileio import File
except ImportError:
    File = Path

if TYPE_CHECKING:
    # from magic import FileMagic
    from fileio import FileLike

InputContentType = TypeVar('InputContentType', Union[bytes, Iterable[Union[bytes, Any]]], File, str, Any)
InputPathType = TypeVar('InputPathType', File, str, Any) # Union[Path, str, Any]
InputSourceType = TypeVar('InputSourceType', File, str, bytes, Any)
InputSourceTupleType = Tuple[int, InputSourceType]
InputSourceT = Union[InputSourceType, InputSourceTupleType]

OutputType = TypeVar('OutputType')

class InvalidTargetError(Exception):
    """
    Invalid Target Error
    """
    def __init__(
        self,
        converter: 'BaseConverter',
        target: str,
    ):
        """
        Initialize the Invalid Target Error
        """
        self.converter = converter
        self.target = target
        super().__init__(f'[{self.converter.name}] Invalid conversion target: {target} (supported: {self.converter.targets})')

class InvalidSourceError(Exception):
    """
    Invalid Source Error
    """
    def __init__(
        self,
        converter: 'BaseConverter',
        source: str,
    ):
        """
        Initialize the Invalid Source Error
        """
        self.converter = converter
        self.source = source
        super().__init__(f'[{self.converter.name}] Invalid conversion source: {source} (supported: {self.converter.source})')


class BaseConverter(abc.ABC):
    """
    The Base Conversion Class
    """
    name: Optional[str] = None
    source: Optional[str] = None # The source format
    targets: Optional[List[str]] = None # List of supported targets
    enabled: Optional[bool] = None # Whether the converter is enabled
    async_enabled: Optional[bool] = None # Whether the converter is async enabled


    def __init__(
        self,
        source: Optional[str] = None,
        targets: Optional[List[str]] = None,
        enabled: Optional[bool] = None,
        async_enabled: Optional[bool] = None,
        **kwargs,
    ):
        """
        Initialize the Converter
        """
        if source is not None: self.source = source
        if targets is not None: self.targets = targets
        if enabled is not None: self.enabled = enabled
        if async_enabled is not None: self.async_enabled = async_enabled
        
        # Remove the leading dot and convert to lowercase
        if self.source: self.source = self.source.lstrip('.').lower()
        if self.targets: self.targets = [t.lstrip('.').lower() for t in self.targets]
        
        self.pre_init(**kwargs)
        if self.enabled is None: self.enabled = self.validate_enabled(**kwargs)
        self.post_init(**kwargs)
        self.finalize_init(**kwargs)
    
    @property
    def logger(self) -> Logger:
        """
        Return the logger
        """
        return default_logger
    
    @property
    def null_logger(self) -> Logger:
        """
        Return the null logger
        """
        return null_logger
    
    @property
    def pool(self) -> Type[ThreadPoolV2]:
        """
        Return the pool
        """
        return ThreadPoolV2

    def pre_init(self, **kwargs):
        """
        Pre-initialize the converter
        """
        pass

    def post_init(self, **kwargs):
        """
        Post-initialize the converter
        """
        pass

    def finalize_init(self, **kwargs):
        """
        Finalize the converter
        """
        pass
    
    @property
    def source_ext(self) -> str:
        """
        Return the source extension
        """
        return f'.{self.source}' if self.source else None
    
    @property
    def source_mime_type(self) -> str:
        """
        Return the source mime type
        """
        return f'application/{self.source}' if self.source else None
    
    @property
    def target_exts(self) -> List[str]:
        """
        Return the target extensions
        """
        return [f'.{t}' for t in self.targets] if self.targets else None
    
    @property
    def target_mime_types(self) -> List[str]:
        """
        Return the target mime types
        """
        return [f'application/{t}' for t in self.targets] if self.targets else None

    def validate_enabled(self, **kwargs) -> bool:
        """
        Validate whether the converter is enabled
        """
        return True

    def is_valid_source_content_type(
        self,
        content_type: Optional[str] = None,
        **kwargs,
    ) -> bool:
        """
        Validate whether the content type is valid

        content-type: application/pdf
        """
        if '.' in content_type: content_type = content_type.split('.')[-1]
        return self.source in content_type.lower() if content_type else False
    
    def is_valid_source_type(
        self,
        data: Optional[InputSourceType] = None,
        **kwargs,
    ) -> bool:
        """
        Validate whether the data is a valid source type for conversion

        data: .pdf
        """
        if data is None: return False
        if isinstance(data, Path) or hasattr(data, 'suffix'): return data.suffix == self.source_ext
        if isinstance(data, str):
            if len(data) < 5:
                return data == self.source_ext if '.' in data else data == self.source
            return self.detect_content_type(data, mime = True) == self.source_mime_type
        return self.is_valid_source_path(data, **kwargs)

    def is_valid_source_path(
        self,
        path: Optional[InputPathType] = None,
        **kwargs,
    ) -> bool:
        """
        Validate whether the path is valid

        path: /path/to/file.pdf
        """
        if path is None: return False
    
        if not isinstance(path, str) and hasattr(path, 'suffix'): return path.suffix == self.source_ext
        return str(path).endswith(self.source_ext)
    
    def is_valid_target_content_type(
        self,
        content_type: Optional[str] = None,
        **kwargs,
    ) -> bool:
        """
        Validate whether the content type is a valid target content type for conversion

        content-type: application/pdf
        """
        if '.' in content_type: content_type = content_type.split('.')[-1]
        return (
            any(t in content_type.lower() for t in self.targets)
            if content_type
            else False
        )
    

    def is_valid_target_type(
        self,
        data: Optional[InputContentType] = None,
        **kwargs,
    ) -> bool:
        """
        Validate whether the data is a valid target type for conversion
        """
        if data is None: return False
        if not isinstance(data, str) and hasattr(data, 'suffix'): return data.suffix in self.target_exts
        if isinstance(data, str):
            if len(data) < 5:
                return data in self.target_exts if '.' in data else data in self.targets
            return self.detect_content_type(data, mime = True) in self.target_mime_types
        return self.is_valid_target_path(data, **kwargs)
    
    def is_valid_target_path(
        self,
        path: Optional[InputPathType] = None,
        **kwargs,
    ) -> bool:
        """
        Validate whether the path is a valid target path for conversion

        path: /path/to/file.pdf
        """
        if path is None: return False
    
        if not isinstance(path, str) and hasattr(path, 'suffix'): return path.suffix in self.target_exts
        return any(str(path).endswith(t) for t in self.target_exts)
    

    def _convert_source_to_target(
        self,
        source: InputSourceT,
        target: str, 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> OutputType:
        """
        Convert the source to the target

        source: /path/to/file.pdf
        target: '.docx'
        """
        raise NotImplementedError
    

    def convert_source_to_target(
        self,
        source: InputSourceType,
        target: str, 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> OutputType:
        """
        Convert the source to the target

        source: /path/to/file.pdf
        target: '.docx'
        """
        if not self.enabled: raise NotImplementedError(f'{self.name} is not enabled')
        return self._convert_source_to_target(
            source=source,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    def _convert_source_to_targets_one(
        self,
        target: Tuple[int, str], 
        source: InputSourceT,
        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,
        **kwargs,
    ) -> Tuple[int, OutputType]:
        """
        Single Process Conversion for Multiple Targets
        """
        index, target = target
        return index, self._aconvert_source_to_target(
            source=source,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    def _convert_source_to_targets(
        self,
        source: InputSourceType,
        targets: List[str], 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> List[OutputType]:
        """
        Convert the source to the targets

        source: /path/to/file.pdf
        targets: ['.docx', '.txt']
        """
        results = {}
        data_iterator = list(enumerate(targets))
        for item in self.pool.sync_iterate(
            self._convert_source_to_targets_one,
            data_iterator,
            source = source,
            source_filename = source_filename,
            target_output = target_output,
            return_ordered = False,
            **kwargs,
        ):
            index, output = item
            results[index] = output
        
        # Reorder the results to match the input order
        return [results[i] for i in range(len(results))]
    
    

    def convert_source_to_targets(
        self,
        source: InputSourceType,
        targets: List[str], 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> Dict[str, OutputType]:
        """
        Convert the source to the targets

        source: /path/to/file.pdf
        targets: ['.docx', '.txt']
        """
        if not self.enabled: raise NotImplementedError(f'{self.name} is not enabled')
        return self._convert_source_to_targets(
            source=source,
            targets=targets,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    def _convert_sources_to_target_one(
        self,
        source: Tuple[int, InputSourceType],
        target: str, 
        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,
        **kwargs,
    ) -> Tuple[int, OutputType]:
        """
        Single Process Conversion for Multiple Sources
        """
        index, source = source
        return index, self._convert_source_to_target(
            source=source,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    def _convert_sources_to_target(
        self,
        sources: Iterable[InputSourceType],
        target: str,
        
        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> List[OutputType]:
        """
        Convert the sources to the target

        sources: ['/path/to/file.pdf', '/path/to/file2.pdf']
        target: '.docx'
        """
        results = {}
        data_iterator = list(enumerate(sources))
        for item in self.pool.sync_iterate(
            self._convert_sources_to_target_one,
            data_iterator,
            target = target,
            source_filename = source_filename,
            target_output = target_output,
            return_ordered = False,
            **kwargs,
        ):
            index, output = item
            results[index] = output

        # Reorder the results to match the input order
        return [results[i] for i in range(len(results))]
    
    def convert_sources_to_target(
        self,
        sources: Iterable[InputSourceType],
        target: str, 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> Dict[str, OutputType]:
        """
        Convert the sources to the target

        sources: ['/path/to/file.pdf', '/path/to/file2.pdf']
        target: '.docx'
        """
        if not self.enabled: raise NotImplementedError(f'{self.name} is not enabled')
        return self._convert_sources_to_target(
            sources=sources,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )


    async def _aconvert_source_to_target(
        self,
        source: InputSourceT,
        target: str, 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,
        **kwargs,
    ) -> OutputType:
        """
        Convert the source to the target

        source: /path/to/file.pdf
        target: '.docx'
        """
        raise NotImplementedError
    

    async def _aconvert_source_to_targets_one(
        self,
        target: Tuple[int, str], 
        source: InputSourceT,
        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,
        **kwargs,
    ) -> Tuple[int, OutputType]:
        """
        Single Process Conversion for Multiple Targets
        """
        index, target = target
        return index, await self._aconvert_source_to_target(
            source=source,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    async def _aconvert_sources_to_target_one(
        self,
        source: Tuple[int, InputSourceType],
        target: str, 
        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,
        **kwargs,
    ) -> Tuple[int, OutputType]:
        """
        Single Process Conversion for Multiple Sources
        """
        index, source = source
        return index, await self._aconvert_source_to_target(
            source=source,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    async def aconvert_source_to_target(
        self,
        source: InputSourceType,
        target: str, 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> OutputType:
        """
        Convert the source to the target

        source: /path/to/file.pdf
        target: '.docx'
        """
        if not self.enabled: raise NotImplementedError(f'{self.name} is not enabled')
        return await self._aconvert_source_to_target(
            source=source,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    

    async def _aconvert_source_to_targets(
        self,
        source: InputSourceType,
        targets: List[str], 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> List[OutputType]:
        """
        Convert the source to the targets

        source: /path/to/file.pdf
        targets: ['.docx', '.txt']
        """
        results = {}
        data_iterator = list(enumerate(targets))
        async for item in self.pool.async_iterate(
            self._aconvert_source_to_targets_one,
            data_iterator,
            source = source,
            source_filename = source_filename,
            target_output = target_output,
            return_ordered = False,
            **kwargs,
        ):
            index, output = item
            results[index] = output

        # Reorder the results to match the input order
        return [results[i] for i in range(len(results))]
    

    async def aconvert_source_to_targets(
        self,
        source: InputSourceType,
        targets: List[str], 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> Dict[str, OutputType]:
        """
        Convert the source to the targets

        source: /path/to/file.pdf
        targets: ['.docx', '.txt']
        """
        if not self.enabled: raise NotImplementedError(f'{self.name} is not enabled')
        return await self._aconvert_source_to_targets(
            source=source,
            targets=targets,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    async def _aconvert_sources_to_target(
        self,
        sources: Iterable[InputSourceType],
        target: str, 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> List[OutputType]:
        """
        Convert the sources to the target

        sources: ['/path/to/file.pdf', '/path/to/file2.pdf']
        target: '.docx'
        """
        results = {}
        data_iterator = list(enumerate(sources))
        async for item in self.pool.async_iterate(
            self._aconvert_sources_to_target_one,
            data_iterator,
            target = target,
            source_filename = source_filename,
            target_output = target_output,
            return_ordered = False,
            **kwargs,
        ):
            index, output = item
            results[index] = output

        # Reorder the results to match the input order
        return [results[i] for i in range(len(results))]
    
    async def aconvert_sources_to_target(
        self,
        sources: Iterable[InputSourceType],
        target: str, 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        **kwargs,
    ) -> Dict[str, OutputType]:
        """
        Convert the sources to the target

        sources: ['/path/to/file.pdf', '/path/to/file2.pdf']
        target: '.docx'
        """
        if not self.enabled: raise NotImplementedError(f'{self.name} is not enabled')
        return await self._aconvert_sources_to_target(
            sources=sources,
            target=target,
            source_filename=source_filename,
            target_output=target_output,
            **kwargs,
        )
    
    """
    Misc. Utilities
    """

    @classmethod
    def convert_bytes_to_text(cls, data: bytes, encoding: Optional[str] = 'utf-8', errors: Optional[str] = 'ignore', **kwargs) -> str:
        """
        Converts file bytes to text
        """
        return data.decode(encoding, errors=errors)
    
    @classmethod
    def convert_text_to_bytes(cls, data: str, encoding: Optional[str] = 'utf-8', errors: Optional[str] = 'ignore', **kwargs) -> bytes:
        """
        Converts file text to bytes
        """
        return data.encode(encoding, errors=errors)
    
    @classmethod
    def convert_bytes_to_file(cls, data: bytes, path: Optional[InputPathType] = None, make_temp: Optional[bool] = False, **kwargs) -> 'FileLike':
        """
        Converts file bytes to a file
        """
        path = File(path) if path and not make_temp else File(tempfile.mktemp())
        path.write_bytes(data)
        return path
    
    @classmethod
    def convert_text_to_file(cls, data: str, path: Optional[InputPathType] = None, make_temp: Optional[bool] = False, **kwargs) -> 'FileLike':
        """
        Converts file text to a file
        """
        path = File(path) if path and not make_temp else File(tempfile.mktemp())
        path.write_text(data)
        return path
    
    @classmethod
    def convert_file_to_bytes(cls, path: InputPathType, **kwargs) -> bytes:
        """
        Converts a file to bytes
        """
        return Path(path).read_bytes()
    
    @classmethod
    def convert_file_input_to_bytes(
        cls, 
        data: InputContentType, 
        **kwargs
    ) -> bytes:
        """
        Converts input to bytes
        """
        if isinstance(data, bytes): return data
        if isinstance(data, str): return Path(data).read_bytes()
        if isinstance(data, Iterable): return b''.join(data)
        if isinstance(data, Path) or hasattr(data, 'read_bytes'): return data.read_bytes()
        raise TypeError(f'Invalid data type: {type(data)}')
    
    @classmethod
    def convert_file_input_to_file(
        cls, 
        data: InputContentType, 
        path: Optional[InputPathType] = None, 
        make_temp: Optional[bool] = False,
        **kwargs
    ) -> 'FileLike':
        """
        Converts input to a file
        """
        if isinstance(data, bytes): 
            return cls.convert_bytes_to_file(data, path=path, make_temp=make_temp)
        if isinstance(data, Path) or hasattr(data, 'as_posix'): 
            return cls.convert_bytes_to_file(data.read_bytes()) if make_temp else data
        if isinstance(data, str):
            # Try to validate it as a path
            p = File(data)
            if p.exists():
                return cls.convert_bytes_to_file(p.read_bytes()) if make_temp else p
            return cls.convert_text_to_file(data, path=path, make_temp=make_temp)
        if isinstance(data, Iterable): return cls.convert_bytes_to_file(b''.join(data), path=path, make_temp=make_temp)
        raise TypeError(f'Invalid data type: {type(data)}')

    
    @classmethod
    def detect_content_type(
        cls, 
        data: InputContentType,
        mime: Optional[bool] = False,
        **kwargs
    ) -> str:
        """
        Detect the content type of the data
        """
        from lazyops.imports._filemagic import resolve_magic
        resolve_magic(True)
        from magic import from_file, from_buffer
        if isinstance(data, str): return from_file(data, mime=mime)
        if isinstance(data, bytes): return from_buffer(data, mime = mime)
        return from_buffer(cls.convert_file_input_to_bytes(data), mime = mime)

    @classmethod
    def create_hash(cls, data: Union[bytes, Iterable[Union[bytes, Any]]]) -> str:
        """
        Create a hash for the data
        """
        if isinstance(data, bytes): return hash_func(data).hexdigest()
        return hash_func(b''.join(data)).hexdigest()
    
    @classmethod
    async def acreate_hash(cls, data: Union[bytes, Iterable[Union[bytes, Any]]]) -> str:
        """
        Create a hash for the data
        """
        return await ThreadPoolV2.run_async(cls.create_hash, data)


from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Sequence

from lazyops.libs.dbinit.data_structures.privileges import Privilege

if TYPE_CHECKING:
    from lazyops.libs.dbinit.entities.role import Role


@dataclass
class GrantTo:
    """
    Represents a Sequence of :class:`lazyops.libs.dbinit.data_structures.Privilege` to grant to a
    :class:`lazyops.libs.dbinit.entities.Role` for a given :class:`lazyops.libs.dbinit.mixins.Grantable`.
    """

    privileges: Sequence[Privilege]
    to: Sequence[Role]

    def __repr__(self) -> str:
        return f"<GrantTo (privileges={self.privileges}, to={self.to})>"



from __future__ import annotations

"""
Extra Helper Types
"""

import threading

class Singleton(object):
    
    """
    This is a singleton that is not locked to a single thread
    """

    __instance = None

    def __new__(cls):
        """
        If the instance is not created, then create it
        """
        if cls.__instance is None:
            cls.__instance = super(Singleton, cls).__new__(cls)
        return cls.__instance


class LockedSingleton(object):

    """
    This is a singleton that is locked to a single thread
    """
    __instance = None
    __instance_lock = threading.RLock()

    def __new__(cls):
        """
        If the instance is not created, then create it
        """
        if cls.__instance is None:
            with cls.__instance_lock:
                cls.__instance = super(LockedSingleton, cls).__new__(cls)
        return cls.__instance




"""
This is now empty to avoid importing all the submodules when you import lazyops.
"""

# from __future__ import absolute_import

# from lazyops.types.classprops import lazyproperty, classproperty, cached_classproperty
# from lazyops.types import (
#     BaseModel, 
#     BaseSettings, 
#     Field, 
#     validator
# )

# from lazyops.utils import logger
# from lazyops.utils.lazylib import LazyLib
# from lazyops.utils.imports import (
#     resolve_missing,
#     require_missing_wrapper,
# )


"""
ZStd Compress

Note - ZStd doesn't work as well as it returned some errors when decompressing.
"""

from .base import BaseCompress
from typing import Union

try:
    import zstd
    _zstd_available = True
except ImportError:
    _zstd_available = False

def ensure_zstd_available():
    """
    Ensure Zstd is available
    """
    global _zstd_available, zstd
    if _zstd_available is False: 
        from lazyops.utils.imports import resolve_missing
        resolve_missing('zstd', required = True)
        import zstd
        _zstd_available = True
        globals()['zstd'] = zstd


class ZStdCompress(BaseCompress):


    @classmethod
    def compress(cls, data: Union[str, bytes], encoding: str = 'utf-8', **kwargs) -> bytes:
        """
        ZStd Compress
        """
        ensure_zstd_available()
        if isinstance(data, str): data = data.encode(encoding)
        return zstd.compress(data)

    @classmethod
    def decompress(cls, data: Union[str, bytes], encoding: str = 'utf-8', **kwargs) -> bytes:
        """
        ZStd Decompress
        """
        ensure_zstd_available()
        if isinstance(data, str): data = data.encode(encoding = encoding)
        return zstd.decompress(data)




from __future__ import annotations

"""
Lazy Function Wrappers
"""
import functools
from lazyops.libs.pooler import is_coro_func, ThreadPooler
from typing import TypeVar, Callable, Any

WrappedFnReturnT = TypeVar("WrappedFnReturnT")
WrappedFn = TypeVar("WrappedFn", bound = Callable[..., Any])
ReturnT = TypeVar("ReturnT")


def lazy_function_wrapper(
    function: Callable[..., ReturnT],
    *function_args,
    **function_kwargs,
) -> Callable[..., ReturnT]:
    """
    Lazy Function Wrapper

    This function is used to handle lazily initializing a wrapper function (retry, etc)
    The execution of the function is deferred until the first call to the wrapper function

    If the returned function is None after initialization is called with arguments, then
    the initialized function will not be wrapped and the original function will be returned
    """
    _initialized = False
    _initialized_function = None

    def lazywrapped(func: Callable[..., ReturnT]) -> Callable[..., ReturnT]:
        """
        Wrapper Function
        """
        if is_coro_func(func):
            @functools.wraps(func)
            async def _wrapper(*args, **kwargs) -> ReturnT:
                """
                Wrapped Function
                """
                nonlocal _initialized_function, _initialized
                if not _initialized:
                    _initialized_function = await ThreadPooler.asyncish(function, *function_args, **function_kwargs)
                    _initialized = True
                if _initialized_function is None:
                    return await func(*args, **kwargs)
                return await _initialized_function(func)(*args, **kwargs)
            return _wrapper
        

        @functools.wraps(func)
        def _wrapper(*args, **kwargs) -> ReturnT:
            """
            Wrapped Function
            """
            nonlocal _initialized_function, _initialized
            if not _initialized:
                _initialized_function = function(*function_args, **function_kwargs)
                _initialized = True

            if _initialized_function is None:
                # If the function is None, return the original function
                return func(*args, **kwargs)
            
            return _initialized_function(func)(*args, **kwargs)
        return _wrapper
    
    return lazywrapped




from ._base import *
from .types import CreateSchemaType, UpdateSchemaType
from .models import LazyHasher, LazyUserSchema, LazyDBConfig, LazyDBSaveMetrics


logger = get_logger('LazyDB', module='core')


# Creates various lookup values for relationships
class LazyDBIndex:
    pass


class LazyDBModel:
    def __init__(self, name, schema, hash_schema: Dict[str, str] = None, is_dev: bool = False):
        self.name = name
        self.base_schema = schema
        self.hash_schema = hash_schema
        self.is_prod = not is_dev
        if self.is_prod:
            logger.info(f'[{self.index_name} Index] Production Mode Enabled. Removing Non-Hashed Values')
        self.create_schema()
        self.index = {}
        self.lookup = {}
        self.idx = 0
    
    def create_schema(self):
        if isinstance(self.base_schema, str):
            self.base_schema = sjson_loads(self.base_schema)
        assert isinstance(self.base_schema, dict), 'Schema must be a Dict or String that is JSON Decodable'
        self.schema = create_model(self.name, **self.id_schema, **self.base_schema)
        self.schema_props = list(self.schema.schema()['properties'].keys())
        setattr(sys.modules[self.schema.__module__], self.schema.__name__, self.schema)

    @timed_cache(15, 512)
    def validate_idx(self, uid: int = None, dbid: str = None, *args, **kwargs):
        if uid is not None and self.index.get(uid):
            return uid
        if dbid is not None and self.lookup.get(dbid):
            return self.lookup[dbid]
        return None
    
    async def async_validate_idx(self, uid: int = None, dbid: str = None, *args, **kwargs):
        if uid is not None and self.index.get(uid):
            return uid
        if dbid is not None and self.lookup.get(dbid):
            return self.lookup[dbid]
        return None

    @timed_cache(5)
    def get_by_id(self, uid: int = None, dbid: str = None, *args, **kwargs):
        idx = self.validate_idx(uid=uid, dbid=dbid, *args, **kwargs)
        if idx is None:
            return None
        return self.index.get(idx)
    
    @timed_cache(15)
    def match_props(self, item, name, val, *args, **kwargs):
        item_data = jsonable_encoder(item)
        if not item_data.get(name):
            return False
        by_type = kwargs.get('ByType', False)
        if by_type:
            return isinstance(item_data[name], val)
        # Prop in list
        if isinstance(val, list) and isinstance(item_data[name], str):
            return bool(item_data[name] in val)
        # val in item's list
        if isinstance(val, str) and isinstance(item_data[name], list):
            return bool(val in item_data[name])
        # not NoneType
        if isinstance(val, str) and val == 'NotNone':
            return bool(item_data[name] is not None)
        if isinstance(val, str) and val == 'NotNoneType':
            return isinstance(item_data[name], NoneType)
        return item_data[name] == val

    @timed_cache(5)
    def get_by_props(self, props: Dict[str, Any], forward=True, *args, **kwargs):
        if not props:
            return None
        search_range = range(self.idx) if forward else range(self.idx, 0, -1)
        for idx in search_range:
            if self.index.get(idx):
                item = self.index[idx]
                for name, val in props.items():
                    if name not in self.schema_props:
                        continue
                    if self.match_props(item, name, val, *args, **kwargs):
                        return item
        
        return None

    @timed_cache(5)
    def get(self, uid: int = None, dbid: str = None, props: Dict[str, Any] = None, *args, **kwargs):
        logger.info(f'[{self.index_name} Index]: GET Request for UID: {uid}: DBID: {dbid}, Props: {props}, args: {args}, kwargs: {kwargs}')
        return self.get_by_props(props=props, *args, **kwargs) if props is not None else self.get_by_id(uid=uid, dbid=dbid, *args, **kwargs)
    
    async def async_get(self, uid: int = None, dbid: str = None, props: Dict[str, Any] = None, *args, **kwargs):
        return self.get_by_props(props=props, *args, **kwargs) if props else self.get_by_id(uid=uid, dbid=dbid, *args, **kwargs)

    def filter_items(self, item_list: List[Any], props: Dict[str, Any] = None, *args, **kwargs):
        res_list = []
        for item in item_list:
            for name, val in props.items():
                if name not in self.schema_props:
                    continue
                if self.match_props(item, name, val, *args, **kwargs):
                    res_list.append(item)
        return res_list

    def get_many(self, id_list: List[int] = None, dbid_list: List[int] = None, props: Dict[str, Any] = None, *args, **kwargs):
        if not id_list and not dbid_list:
            return None
        get_res = [self.get(uid=idx) for idx in id_list] if id_list else [self.get(dbid=dbid) for dbid in dbid_list]
        get_res = [i for i in get_res if i]
        if props:
            get_res = self.filter_items(get_res, props, *args, **kwargs)
        return get_res
    
    async def async_get_many(self, id_list: List[int] = None, dbid_list: List[int] = None, props: Dict[str, Any] = None, *args, **kwargs):
        if not id_list and not dbid_list:
            return None
        get_list = id_list or dbid_list
        tasks = [asyncio.ensure_future(self.async_get(uid=idx)) for idx in get_list]
        all_tasks = await asyncio.gather(*tasks)
        get_res = [i for i in all_tasks if i]
        if props:
            get_res = self.filter_items(get_res, props, *args, **kwargs)
        return get_res

    def create(self, data, *args, **kwargs):
        data = self.create_or_update_hash(data)
        new_item = self.schema(uid=self.get_idx(), dbid=self.get_dbid(), *args, **data, **kwargs)
        self.index[self.idx] = new_item
        self.lookup[self.current_id] = self.idx
        self.idx += 1
        return new_item
    
    async def async_create(self, data, *args, **kwargs):
        data = self.create_or_update_hash(data)
        new_item = self.schema(uid=self.get_idx(), dbid=self.get_dbid(), *args, **data, **kwargs)
        self.index[self.idx] = new_item
        self.lookup[self.current_id] = self.idx
        self.idx += 1
        return new_item
        
    def remove(self, uid: int = None, dbid: str = None, *args, **kwargs):
        rm_id = self.validate_idx(uid=uid, dbid=dbid, *args, **kwargs)
        if not rm_id:
            return None
        item = self.index.pop(rm_id, None)
        _ = self.lookup.pop(item.dbid)
        return rm_id
    
    async def async_remove(self, uid: int = None, dbid: str = None, *args, **kwargs):
        rm_id = await self.async_validate_idx(uid=uid, dbid=dbid, *args, **kwargs)
        if not rm_id:
            return None
        item = self.index.pop(rm_id, None)
        _ = self.lookup.pop(item.dbid)
        return rm_id
    
    def update(self, data: Union[UpdateSchemaType, Dict[str, Any]], uid: int = None, dbid: str = None, prop_name: str = None, prop_val: Any = None, *args, **kwargs):
        item = self.get(id, dbid, prop_name=prop_name, prop_val=prop_val, *args, **kwargs)
        if not item:
            return None
        item_data = jsonable_encoder(item)
        update_data = data if isinstance(data, dict) else data.dict(exclude_unset=True)
        item, update_data = self.create_or_update_hash(update_data=update_data, item=item)
        for field in item_data:
            if field in update_data:
                setattr(item, field, update_data[field])
        idx = item.uid
        item.updated = self.get_timestamp()
        self.index[idx] = item
        return self.index[idx]
    
    async def async_update(self, data: Union[UpdateSchemaType, Dict[str, Any]], uid: int = None, dbid: str = None, prop_name: str = None, prop_val: Any = None, *args, **kwargs):
        item = await self.async_get(uid, dbid, prop_name=prop_name, prop_val=prop_val, *args, **kwargs)
        if not item:
            return None
        item_data = jsonable_encoder(item)
        update_data = data if isinstance(data, dict) else data.dict(exclude_unset=True)
        item, update_data = self.create_or_update_hash(update_data=update_data, item=item)
        for field in item_data:
            if field in update_data:
                setattr(item, field, update_data[field])
        idx = item.uid
        item.updated = self.get_timestamp()
        self.index[idx] = item
        return self.index[idx]

    def clear_caches(self):
        raise NotImplementedError

    @cached_property
    def model_name(self):
        return self.name.capitalize()
    
    @cached_property
    def logger_name(self):
        return f'[{self.model_name} Model]'

    @property
    def id_schema(self):
        return {
            'uid': (int, Field(default_factory=self.get_idx)),
            'dbid': (str, Field(default_factory=self.get_dbid)),
            'created': (str, Field(default_factory=self.get_timestamp)),
            'updated': (str, Field(default_factory=self.get_timestamp))
        }

    @property
    def current_id(self):
        return f'{self.name}_{self.idx}'
    
    @property
    def has_hashing(self):
        return bool(self.hash_schema)
    
    def get_idx(self):
        return self.idx
    
    def get_dbid(self):
        return self.current_id
    
    def get_timestamp(self):
        return tstamp()
    
    def create_or_update_hash(self, data=None, update_data=None, item=None):
        if not self.has_hashing:
            if data:
                return data
            return item, update_data
        logger.info(self.hash_schema)
        for field, hash_field in self.hash_schema.items():
            if data and data.get(field):
                data[hash_field] = self.create_hash(val=data[field])
                if self.is_prod:
                    _ = data.pop(field)
            elif update_data and update_data.get(field):
                item, updated = self.update_hash(prop=field, hash_prop=hash_field, new_val=update_data[field], do_verify=True, item=item)
                if updated:
                    logger.info(f'[{self.index_name} Index]: Updated Hashed Field: {field} = {hash_field}')
                    if not self.is_prod:
                        setattr(item, field, update_data[field])
                else:
                    logger.error(f'[{self.index_name} Index]: Failed Validation for Hash Field {field} = {hash_field}')
                _ = update_data.pop(field)
        if data:
            return data
        return item, update_data

    
    @classmethod
    def create_hash(cls, prop=None, hash_prop=None, item=None, item_data=None, val=None, *args, **kw):
        if item is None and item_data is None and val is None:
            raise ValueError
        if val:
            return LazyHasher.create(val)
        item_data = item_data or jsonable_encoder(item)
        hash_results = LazyHasher.create(item_data[prop])
        if not item:
            return {hash_prop: hash_results}
        setattr(item, hash_prop, LazyHasher.create[item_data[prop]])
        return item
    
    @classmethod
    def update_hash(cls, prop, hash_prop, new_val, do_verify=True, item=None, item_data=None, *args, **kw):
        if item is None and item_data is None:
            raise ValueError
        item_data = item_data or jsonable_encoder(item)
        hash_results = LazyHasher.update(item_data[hash_prop], item_data[prop], new_val, do_verify=do_verify)
        if not item:
            return {'updated': bool(hash_results), hash_prop: hash_results}
        if not hash_results:
            return item, False
        setattr(item, hash_prop, hash_results)
        return item, True
    
    def __call__(self, method, uid: int = None, dbid: str = None, *args, **kwargs):
        func = getattr(self, method)
        return func(uid=uid, dbid=dbid, **kwargs)

# Associates Relationships between object classes
class LazyDBRelations:
    def __init__(self, name, method):
        pass

class LazyDBBase:
    def __init__(self, dbcache: Any, config: LazyDBConfig):
        self.config = config
        self.cache = dbcache
        self.alive = True
        self.lock = threading.RLock()
        self.setup_db_schema()
        self.init_db()
        self.migrate_db()
        self.finalize_init()
    
    def init_db(self):
        if self.cache.exists: 
            dbdata = self.cache.restore()
            self._db = dbdata['db']
            self._alivetime = dbdata['timer']
            self.metrics = dbdata['metrics']
            self.metrics.num_loaded += 1
            self.metrics.last_load = tstamp()
        else:
            self._alivetime = timer()
            self.metrics = LazyDBSaveMetrics(created=tstamp())
    
    def setup_db_schema(self):
        self._db = {}
        self._base_schemas = {}
        self.__class__.__name__ = 'LazyDB' if not self.config.dbname else f'{self.config.dbname}_LazyDB'
        logger.info(f'Initializing {self.__class__.__name__}')
        logger.info(f'[{self.dbname} Setup]: Setting Up DB Schema')
        if self.config.autouser:
            logger.info(f'[{self.dbname} Setup]: Creating Auto User Schema(s)')
            if self.config.userconfigs:
                for name, schema_config in self.config.userconfigs.items():
                    schema = LazyUserSchema.get_schema(schema_config, is_dev=self.config.is_dev)
                    self._db[name] = LazyDBModel(name, schema, LazyUserSchema.get_hash_schema(), is_dev=self.config.is_dev)
                    self._base_schemas[name] = self._db[name].schema
                    logger.info(f'[{self.dbname} Setup]: Created Custom User Schema: {name}')
            else:
                schema = LazyUserSchema.get_schema(is_dev=self.config.is_dev)
                self._db['user'] = LazyDBModel('user', schema, LazyUserSchema.get_hash_schema(), is_dev=self.config.is_dev)
                self._base_schemas['user'] = self._db['user'].schema
                logger.info(f'[{self.dbname} Setup]: Created Default User Schema')

        for name, schema in self.config.dbschema.items():
            hashschema = self.config.hashschema.get(name, None) if self.config.hashschema else None
            self._db[name] = LazyDBModel(name, schema, hashschema, is_dev=self.config.is_dev)
            self._base_schemas[name] = self._db[name].schema
            logger.info(f'[{self.dbname} Setup]: Created DB Schema Added for: {name}')
    
    def migrate_db(self):
        if not self.config.seeddata:
            logger.info(f'[{self.dbname} Migrate]: No Seed Data Provided. Skipping Migration')
        for schema_name, index in  self._db.items():
            if index.idx == 0 and self.config.seeddata.get(schema_name):
                logger.info(f'[DB Migrate]: Running Migration for {schema_name}')
                for item in self.config.seeddata[schema_name]:
                    i = self._db[schema_name].create(data=item)
                    logger.info(f'[{self.dbname} Migrate]: Created Item [{schema_name}] = ID: {i.uid}, DBID: {i.dbid}')

        logger.info(f'[{self.dbname} Migrate]: Completed all Setup and Migration Tasks')
        self.save_db()

    def save_db(self):
        self.metrics.last_save = tstamp()
        self.metrics.num_saved += 1
        self.metrics.time_alive = self._alivetime.ablstime
        dbdata = {'db': self._db, 'timer': self._alivetime, 'metrics': self.metrics}
        self.cache.save(dbdata)


    # Create a Direct Accessible attribute
    def finalize_init(self):
        self.db = LazyObject(self._db)
        for name, index in self._db.items():
            schema_name = f'{name}_schema'
            setattr(self, schema_name, self._base_schemas[name])
            for method_type in ['get', 'create', 'remove', 'update']:
                method_name = f'{name}_{method_type}'
                method_func = (lambda method_index=name, method_type=method_type, *args, **kwargs: self._db[method_index](method=method_type, *args, **kwargs))
                setattr(self, method_name, method_func)

        self.env = LazyEnv
        if self.config.autosave:
            if self.env.is_threadsafe:
                self.start_background()
            else:
                logger.warn(f'[{self.dbname} Finalize]: Currently not threadsafe. Call .start_background in main thread.')

    @property
    def dbname(self):
        return self.__class__.__name__
    
    def validate_cls(self, clsname):
        return self._db.get(clsname, None)

    def get(self, clsname, *args, **kwargs):
        if not self.validate_cls(clsname):
            return None
        return self._db[clsname].get(*args, **kwargs)

    def create(self, clsname, *args, **kwargs):
        if not self.validate_cls(clsname):
            return None
        return self._db[clsname].create(*args, **kwargs)

    def update(self, clsname, *args, **kwargs):
        if not self.validate_cls(clsname):
            return None
        return self._db[clsname].update(*args, **kwargs)
    
    def remove(self, clsname, *args, **kwargs):
        if not self.validate_cls(clsname):
            return None
        return self._db[clsname].remove(*args, **kwargs)
    
    def start_background(self):
        self.env.enable_watcher()
        self.t = threading.Thread(target=self.background, daemon=True)
        self.t.start()
        self.env.add_thread(self.t)
    
    def background(self):
        logger.info(f'[{self.dbname} Background]: DB AutoSaver Active. Saving Every: {self.config.savefreq} secs')
        microsleep = self.config.savefreq / 20
        while self.alive:
            for _ in range(20):
                time.sleep(microsleep)
                if self.env.killed:
                    self.alive = False
                    break
            
            self.save_db()
            if not self.alive:
                break

    def __call__(self, clsname, method, uid: int = None, dbid: str = None, *args, **kwargs):
        func = getattr(self, method)
        return func(self, clsname, uid=uid, dbid=dbid, *args, **kwargs)


class LazyDB(LazyDBBase):
    def __init__(self, dbcache: Any, config: LazyDBConfig):
        super().__init__(dbcache, config)
    




"""
Migrated from lazyops.utils.pooler to allow for non-dependent 3rd party libraries
"""

from .main import (
    set_concurrency_limit,
    get_concurrency_limit,
    is_coro_func,
    amap_iterable,
    ThreadPool,
    ThreadPoolV2,
    ThreadPooler,
    ensure_coro,
)


from __future__ import annotations

"""
Base Persistence Backend
"""
import contextlib
import collections.abc
from pathlib import Path
from pydantic import BaseModel
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Iterable, List, Type, ItemsView, TYPE_CHECKING
from lazyops.utils.helpers import  create_unique_id
from lazyops.utils.logs import logger
# from lazyops.utils.pooler import ThreadPoolV2 as ThreadPooler
from lazyops.utils.pooler import ThreadPooler
from ..serializers import get_serializer, SerializerT
from ..serializers.base import create_obj_hash

if TYPE_CHECKING:
    from lazyops.types.models import BaseSettings
    from ..serializers.base import ObjectValue

SchemaType = TypeVar("SchemaType", bound=BaseModel)


class BaseStatefulBackend(collections.abc.MutableMapping):
    """
    Base Stateful Backend
    """
    serializer: Optional[SerializerT] = None
    serializer_type: Optional[str] = None
    serializer_kwargs: Optional[Dict[str, Any]] = {}

    def __init__(
        self,
        serializer: Optional[str] = None,
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        base_key: Optional[str] = None,
        async_enabled: Optional[bool] = False,
        settings: Optional['BaseSettings'] = None,
        **kwargs,
    ):
        """
        Initializes the backend
        """
        self.base_key = base_key
        self.async_enabled = async_enabled
        self.settings = settings
        if serializer is not None:
            self.serializer_type = serializer
        if serializer_kwargs is not None:
            self.serializer_kwargs = serializer_kwargs
        if self.serializer is None:
            self.serializer = get_serializer(self.serializer_type, **self.serializer_kwargs)
        self._kwargs = kwargs
    
    def get_key(self, key: str) -> str:
        """
        Gets a Key
        """
        return f'{self.base_key}.{key}' if self.base_key else key

    def encode_value(self, value: Union[Any, SchemaType], **kwargs) -> Union[str, bytes]:
        """
        Encodes a Value
        """
        return self.serializer.encode(value, **kwargs)
    
    async def aencode_value(self, value: Union[Any, SchemaType], **kwargs) -> Union[str, bytes]:
        """
        Encodes a Value
        """
        return await self.serializer.aencode(value, **kwargs)

    def decode_value(self, value: Union[str, bytes], **kwargs) -> Any:
        """
        Decodes a Value
        """
        return self.serializer.decode(value, **kwargs)
    
    async def adecode_value(self, value: Union[str, bytes], **kwargs) -> Any:
        """
        Decodes a Value
        """
        return await self.serializer.adecode(value, **kwargs)
    
    def create_hash(self, obj: 'ObjectValue') -> str:
        """
        Creates a Hash
        """
        return self.serializer.create_hash(obj) if self.serializer is not None else \
            create_obj_hash(obj)

    
    def _precheck(self, **kwargs):
        """
        Run a precheck operation
        """
        pass

    def get(self, key: str, default: Optional[Any] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        raise NotImplementedError
            
    def get_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the DB
        """
        raise NotImplementedError


    def set(self, key: str, value: Any, ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        raise NotImplementedError


    def set_batch(self, data: Dict[str, Any], ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        raise NotImplementedError


    def delete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        raise NotImplementedError


    def clear(self, *keys: str, **kwargs):
        """
        Clears the Cache
        """
        raise NotImplementedError

    
    async def aget(self, key: str, default: Optional[Any] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        return await ThreadPooler.run_async(self.get, key, default, **kwargs)

    async def aget_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the DB
        """
        return await ThreadPooler.run_async(self.get_values, keys)

    
    async def aset(self, key: str, value: Any, ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        return await ThreadPooler.run_async(self.set, key, value, ex, **kwargs)


    async def aset_batch(self, data: Dict[str, Any], ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        return await ThreadPooler.run_async(self.set_batch, data, ex, **kwargs)

    async def adelete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        return await ThreadPooler.run_async(self.delete, key, **kwargs)

    async def aclear(self, *keys: str, **kwargs):
        """
        Clears the Cache
        """
        return await ThreadPooler.run_async(self.clear, *keys, **kwargs)

    def iterate(self, **kwargs) -> Iterable[Any]:
        """
        Iterates over the Cache
        """
        raise NotImplementedError


    def get_all_data(self, exclude_base_key: Optional[bool] = False) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        raise NotImplementedError


    def get_all_keys(self, exclude_base_key: Optional[bool] = False) -> List[str]:
        """
        Returns all the Keys
        """
        raise NotImplementedError


    def get_all_values(self) -> List[Any]:
        """
        Returns all the Values
        """
        raise NotImplementedError


    async def aget_all_data(self, exclude_base_key: Optional[bool] = False) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        return await ThreadPooler.run_async(self.get_all_data)

    async def aget_all_keys(self, exclude_base_key: Optional[bool] = False) -> List[str]:
        """
        Returns all the Keys
        """
        return await ThreadPooler.run_async(self.get_all_keys)

    async def aget_all_values(self) -> List[Any]:
        """
        Returns all the Values
        """
        return await ThreadPooler.run_async(self.get_all_values)


    def pop(self, key, default = None):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        value = self.get(key)
        if self.async_enabled:
            ThreadPooler.create_background_task(self.adelete(key))
        else:
            self.delete(key)
        return value if value is not None else default
    
    async def apop(self, key, default = None):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        value = await self.aget(key)
        await self.adelete(key)
        return value if value is not None else default

    def contains(self, key: str) -> bool:
        """
        Returns True if the Cache contains the Key
        """
        raise NotImplementedError

    async def acontains(self, key: str) -> bool:
        """
        Returns True if the Cache contains the Key
        """
        return await ThreadPooler.run_async(self.contains, key)

    def keys(self) -> Iterable[str]:
        """
        Returns the Keys
        """
        return iter(self.get_all_keys(True))

    def get_keys(self, pattern: str, exclude_base_key: Optional[bool] = None) -> List[str]:
        """
        Returns the Keys
        """
        raise NotImplementedError
    
    def values(self) -> Iterable[Any]:
        """
        Returns the Values
        """
        return iter(self.get_all_values())
    
    def items(self, iterable: Optional[bool] = True) -> Union[Iterable[ItemsView], Dict[str, Any]]:
        """
        Returns the Items
        """
        items = self.get_all_data(True)
        return items.items() if iterable else items
        # return iter(self.get_all_data())
    
    async def akeys(self) -> Iterable[Any]:
        """
        Returns the Keys
        """
        return iter(await self.aget_all_keys(True))
    
    async def aget_keys(self, pattern: str, exclude_base_key: Optional[bool] = None) -> List[str]:
        """
        Returns the Keys
        """
        raise NotImplementedError
    
    async def avalues(self) -> Iterable[Any]:
        """
        Returns the Values
        """
        return iter(await self.aget_all_values())
    
    async def aitems(self, iterable: Optional[bool] = True) -> Union[Iterable[ItemsView], Dict[str, Any]]:
        """
        Returns the Items
        """
        items = await self.aget_all_data(True)
        return items.items() if iterable else items
    
    def expire(self, key: str, ex: Optional[int] = None, **kwargs) -> None:
        """
        Expires a Key
        """
        pass
            
    async def aexpire(self, key: str, ex: Optional[int] = None, **kwargs) -> None:
        """
        Expires a Key
        """
        pass
    
    def setdefault(self, key: str, default: Any = None, update_values: Optional[bool] = False, enforce_type: Optional[bool] = False, **kwargs):
        """
        Sets a Default Value
        """
        if self.contains(key):
            with contextlib.suppress(Exception):
                value = self.get(key)
                if value is not None:
                    if enforce_type and not isinstance(value, type(default)):
                        value = type(default)(value)
                    if update_values and isinstance(value, dict) and default and isinstance(default, dict):
                        for k, v in default.items():
                            if k not in value or value[k] is None:
                                
                                value[k] = v
                        self.set(key, value)
                    return value
        self.set(key, default)
        return default
    
    async def asetdefault(self, key: str, default: Any = None, update_values: Optional[bool] = False, enforce_type: Optional[bool] = False, **kwargs):
        """
        Sets a Default Value
        """
        if await self.acontains(key):
            with contextlib.suppress(Exception):
                value = await self.aget(key)
                if value is not None:
                    if enforce_type and not isinstance(value, type(default)):
                        value = type(default)(value)
                    if update_values and isinstance(value, dict) and default and isinstance(default, dict):
                        for k, v in default.items():
                            if k not in value or value[k] is None:
                                value[k] = v
                        await self.aset(key, value)
                    return value
        await self.aset(key, default)
        return default
    
    def update(self, data: Dict[str, Any]):
        """
        Updates the Cache
        """
        for key, value in data.items():
            self.set(key, value)
    
    async def aupdate(self, data: Dict[str, Any]):
        """
        Updates the Cache
        """
        for key, value in data.items():
            await self.aset(key, value)

    def popitem(self):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        key = next(iter(self))
        value = self.get(key)
        if value is not None:
            if self.async_enabled:
                ThreadPooler.create_background_task(self.adelete(key))
            else:
                self.delete(key)
            return key, value
        return None, None
    
    async def apopitem(self):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        key = next(iter(self))
        value = await self.aget(key)
        if value is not None:
            await self.adelete(key)
            return key, value
        return None, None


    def __getitem__(self, key):
        """
        Gets an Item from the DB
        """
        
        return self.get(key)

    def __setitem__(self, key, value):
        """
        Sets an Item in the Cache
        """
        return ThreadPooler.create_background_task(self.aset(key, value)) if \
            self.async_enabled else self.set(key, value)
    
    def __delitem__(self, key):
        """
        Deletes an Item from the Cache
        """
        return ThreadPooler.create_background_task(self.adelete(key)) if \
            self.async_enabled else self.delete(key)

    def __iter__(self):
        """
        Iterates over the Cache
        """
        return self.iterate()
    

    def __len__(self):
        """
        Returns the Length of the Cache
        """
        self._precheck()
        return len(list(self.iterate()))
    

    def __contains__(self, key):
        """
        Returns True if the Cache contains the Key
        """
        return self.contains(key)
    
    def __eq__(self, other):
        """
        Returns True if the Cache is Equal to the Other
        """
        return self.get_all_data(True) == other
    
    def __repr__(self):
        """
        Returns the Representation of the Cache
        """
        return f"<{self.__class__.__name__} num_keys={len(self)}, base_key={self.base_key}, serializer={self.serializer.name}>"
    
    def migrate_compression(self, **kwargs):
        """
        Migrates the Compression
        """
        pass

    async def amigrate_compression(self, **kwargs):
        """
        Migrates the Compression
        """
        pass



"""
Import Handler for `xxhash`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import xxhash
    _xxhash_available = True
except ImportError:
    xxhash = object
    _xxhash_available = False

def resolve_xxhash(
    required: bool = False,
):
    """
    Ensures that `xxhash` is available
    """
    global xxhash, _xxhash_available
    if not _xxhash_available:
        resolve_missing('xxhash', required = required)
        import xxhash
        _xxhash_available = True
        globals()['xxhash'] = xxhash


def require_xxhash(
    required: bool = False,
):
    """
    Wrapper for `resolve_xxhash` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_xxhash, 
            func = func, 
            required = required
        )
    return decorator